"""CLI entry point for seeding SBI forex rates."""

from __future__ import annotations

from fx_bharat.seeds.populate_sbi_forex import main

if __name__ == "__main__":  # pragma: no cover - thin wrapper
    main()



from datetime import date
from pathlib import Path

import pytest

from fx_bharat.db import sqlite_manager as sqlite_manager_module
from fx_bharat.ingestion.models import ForexRateRecord


def test_sqlite_manager_uses_fallback_when_sqlalchemy_missing(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(sqlite_manager_module, "SQLALCHEMY_AVAILABLE", False)

    manager = sqlite_manager_module.SQLiteManager(tmp_path / "fallback.db")
    assert isinstance(manager._backend, sqlite_manager_module._SQLiteFallbackBackend)

    rows = [
        ForexRateRecord(rate_date=date(2024, 3, 1), currency="USD", rate=83.2),
        ForexRateRecord(rate_date=date(2024, 3, 2), currency="EUR", rate=90.4),
    ]
    manager.insert_rates(rows)
    fetched = manager.fetch_range()

    assert len(fetched) == 2
    assert fetched[0].currency == "USD"

    manager.close()


def test_sqlite_fallback_updates_and_filters(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(sqlite_manager_module, "SQLALCHEMY_AVAILABLE", False)

    manager = sqlite_manager_module.SQLiteManager(tmp_path / "fallback_filters.db")

    first = ForexRateRecord(rate_date=date(2024, 4, 1), currency="USD", rate=80.0)
    updated = ForexRateRecord(rate_date=date(2024, 4, 1), currency="USD", rate=81.5)
    sbi_row = ForexRateRecord(
        rate_date=date(2024, 4, 2),
        currency="EUR",
        rate=90.0,
        source="SBI",
        tt_buy=1.0,
        tt_sell=2.0,
        bill_buy=3.0,
        bill_sell=4.0,
        travel_card_buy=5.0,
        travel_card_sell=6.0,
        cn_buy=7.0,
        cn_sell=8.0,
    )

    initial = manager.insert_rates([first, sbi_row])
    follow_up = manager.insert_rates([updated])

    assert initial.inserted == 2
    assert follow_up.updated == 1

    filtered = manager.fetch_range(start=date(2024, 4, 2), source="SBI")
    assert len(filtered) == 1
    assert filtered[0].currency == "EUR"
    assert filtered[0].tt_sell == 2.0

    manager.close()



from __future__ import annotations

from datetime import date
from pathlib import Path

from fx_bharat.db.sqlite_manager import SQLiteManager
from fx_bharat.seeds.populate_sbi_forex import _iter_pdf_paths, seed_sbi_historical, seed_sbi_today


def test_iter_pdf_paths_filters_by_range(tmp_path: Path) -> None:
    resources = tmp_path / "resources"
    resources.mkdir()
    in_range = resources / "2024-01-15.pdf"
    out_of_range = resources / "2024-02-01.pdf"
    invalid = resources / "not-a-date.pdf"
    in_range.write_text("USD 80 81 82 83 84 85 86 87")
    out_of_range.write_text("USD 1 2 3 4 5 6 7 8")
    invalid.write_text("USD 1 2 3 4 5 6 7 8")

    filtered = list(_iter_pdf_paths(resources, date(2024, 1, 1), date(2024, 1, 31)))

    assert filtered == [in_range]


def test_seed_sbi_forex_downloads_latest_when_enabled(monkeypatch, tmp_path: Path) -> None:
    downloaded = tmp_path / "2024-02-02.pdf"
    downloaded.write_text(
        """
        Date: 02/02/2024
        USD 84.0 85.0 83.9 85.1 83.8 85.2 83.7 85.3
        """
    )

    calls: dict[str, int] = {"fetch": 0}

    class DummyDownloader:
        def fetch_latest(self) -> Path:
            calls["fetch"] += 1
            return downloaded

    monkeypatch.setattr("fx_bharat.seeds.populate_sbi_forex.SBIPDFDownloader", DummyDownloader)

    db_path = tmp_path / "db.sqlite"
    result = seed_sbi_historical(
        db_path=db_path, resource_dir=tmp_path / "resources", download=True
    )

    assert calls["fetch"] == 1
    assert result.inserted == 1
    with SQLiteManager(db_path) as manager:
        rows = manager.fetch_range(source="SBI")
    assert len(rows) == 1
    usd = rows[0]
    assert usd.rate_date == date(2024, 2, 2)
    assert usd.tt_buy == 84.0


def test_seed_sbi_today_downloads_and_persists(monkeypatch, tmp_path: Path) -> None:
    today_pdf = tmp_path / "temp.pdf"
    today_pdf.write_text(
        """
        Date: 01/01/2025
        USD 85.22 86.07 85.15 86.24 85.15 86.24 84.15 86.55
        """
    )

    class DummyDownloader:
        def __init__(self, download_dir: Path | str | None = None) -> None:
            self.download_dir = Path(download_dir) if download_dir else tmp_path

        def fetch_latest(self, destination: Path | None = None) -> Path:
            dest = destination or self.download_dir / "FOREX_CARD_RATES.pdf"
            dest.parent.mkdir(parents=True, exist_ok=True)
            dest.write_bytes(today_pdf.read_bytes())
            return dest

    monkeypatch.setattr("fx_bharat.seeds.populate_sbi_forex.SBIPDFDownloader", DummyDownloader)

    resources_dir = tmp_path / "resources"
    db_path = tmp_path / "db.sqlite"
    result = seed_sbi_today(db_path=db_path, resource_dir=resources_dir)

    assert result.inserted == 1
    stored = resources_dir / "2025" / "1" / "2025-01-01.pdf"
    assert stored.exists()
    with SQLiteManager(db_path) as manager:
        rows = manager.fetch_range(source="SBI")
    assert len(rows) == 1
    assert rows[0].rate_date == date(2025, 1, 1)



"""Abstractions for pluggable ingestion strategies."""

from __future__ import annotations

from datetime import date
from pathlib import Path
from typing import Protocol


class IngestionStrategy(Protocol):
    """Contract for fetching raw forex artifacts.

    Concrete implementations are expected to retrieve data for a date range and
    return a local ``Path`` pointing at the downloaded artifact.
    """

    def fetch(self, start_date: date, end_date: date, *, destination: Path | None = None) -> Path:
        ...  # pragma: no cover - protocol definition


__all__ = ["IngestionStrategy"]



from __future__ import annotations

from datetime import date
from pathlib import Path

import pytest

from fx_bharat.db.relational_backend import RelationalBackend
from fx_bharat.ingestion.models import ForexRateRecord, LmeRateRecord


def test_relational_backend_filters_by_source(tmp_path: Path) -> None:
    db_url = f"sqlite:///{tmp_path / 'relational_filters.db'}"
    backend = RelationalBackend(db_url)
    backend.ensure_schema()
    try:
        rows = [
            ForexRateRecord(rate_date=date(2024, 1, 1), currency="USD", rate=82.5, source="RBI"),
            ForexRateRecord(rate_date=date(2024, 1, 2), currency="EUR", rate=90.1, source="SBI"),
        ]
        backend.insert_rates(rows)
        fetched = backend.fetch_range(source="SBI")
        assert len(fetched) == 1
        assert fetched[0].currency == "EUR"
    finally:
        backend.close()


def test_relational_backend_lme_filters_and_invalid_metal(tmp_path: Path) -> None:
    db_url = f"sqlite:///{tmp_path / 'relational_lme.db'}"
    backend = RelationalBackend(db_url)
    backend.ensure_schema()
    try:
        rows = [
            LmeRateRecord(
                rate_date=date(2024, 2, 1),
                price=8500.0,
                price_3_month=8450.0,
                stock=100,
                metal="COPPER",
            ),
            LmeRateRecord(
                rate_date=date(2024, 2, 2),
                price=8600.0,
                price_3_month=8550.0,
                stock=110,
                metal="COPPER",
            ),
        ]
        backend.insert_lme_rates("COPPER", rows)
        fetched = backend.fetch_lme_range("COPPER", start=date(2024, 2, 2))
        assert len(fetched) == 1
        assert fetched[0].rate_date == date(2024, 2, 2)
        with pytest.raises(ValueError):
            backend.insert_lme_rates("GOLD", rows[:1])
    finally:
        backend.close()



from __future__ import annotations

import pytest

from fx_bharat.ingestion.lme import (
    _coerce_date,
    _normalise_metal,
    _parse_float,
    _parse_int,
    parse_lme_table,
)


def test_lme_parse_requires_tables() -> None:
    with pytest.raises(ValueError):
        parse_lme_table("<html></html>", "copper")


def test_lme_normalise_metal_rejects_invalid() -> None:
    with pytest.raises(ValueError):
        _normalise_metal("gold")


def test_lme_parse_helpers_handle_invalid_values() -> None:
    assert _parse_float(None) is None
    assert _parse_float("N/A") is None
    assert _parse_int("10.5") == 10
    assert _parse_int("bad") is None
    assert _coerce_date(object()) is None


def test_lme_parse_requires_data_rows() -> None:
    html = """
    <table>
        <tr><th>Date</th><th>Cash</th></tr>
    </table>
    """
    with pytest.raises(ValueError):
        parse_lme_table(html, "COPPER")



from __future__ import annotations

from datetime import date

from fx_bharat.db.sqlite_backend import SQLiteBackend
from fx_bharat.ingestion.models import ForexRateRecord, LmeRateRecord


def test_sqlite_backend_roundtrip(tmp_path) -> None:
    backend = SQLiteBackend(db_path=tmp_path / "sqlite_backend.db")
    assert backend.ensure_schema() is None

    rows = [
        ForexRateRecord(rate_date=date(2024, 1, 1), currency="USD", rate=82.5),
        ForexRateRecord(rate_date=date(2024, 1, 2), currency="EUR", rate=90.1, source="SBI"),
    ]
    result = backend.insert_rates(rows)
    assert result.inserted == 2

    fetched = backend.fetch_range()
    assert len(fetched) == 2
    assert {row.source for row in fetched} == {"RBI", "SBI"}

    backend.close()


def test_sqlite_backend_lme_roundtrip(tmp_path) -> None:
    backend = SQLiteBackend(db_path=tmp_path / "sqlite_lme.db")

    rows = [
        LmeRateRecord(
            rate_date=date(2024, 2, 1),
            price=8500.0,
            price_3_month=8450.0,
            stock=100,
            metal="COPPER",
        )
    ]
    result = backend.insert_lme_rates("COPPER", rows)
    assert result.total == 1

    fetched = backend.fetch_lme_range("COPPER")
    assert len(fetched) == 1
    assert fetched[0].stock == 100

    backend.close()



from __future__ import annotations

import pytest

from fx_bharat.db import mongo_backend as mongo_module


def test_mongo_backend_requires_client(monkeypatch) -> None:
    monkeypatch.setattr(mongo_module, "MongoClient", None)
    with pytest.raises(ModuleNotFoundError):
        mongo_module.MongoBackend("mongodb://example.com/", database="fx")


def test_mongo_backend_requires_database_name(monkeypatch) -> None:
    class _NoDefaultClient:
        def __init__(self, url: str) -> None:
            self.url = url

        def get_default_database(self):
            return None

    monkeypatch.setattr(mongo_module, "MongoClient", _NoDefaultClient)
    with pytest.raises(ValueError):
        mongo_module.MongoBackend("mongodb://example.com/")


def test_mongo_backend_rejects_unknown_metal() -> None:
    with pytest.raises(ValueError):
        mongo_module.MongoBackend._normalise_metal("gold")



from __future__ import annotations

from datetime import date
from pathlib import Path

import pytest

from fx_bharat.ingestion.models import ForexRateRecord
from fx_bharat.ingestion.sbi_pdf import SBIPDFParseResult
from fx_bharat.seeds import populate_sbi_forex as sbi_module


def test_seed_sbi_historical_rejects_future_end() -> None:
    tomorrow = date.today().replace(day=min(date.today().day + 1, 28))
    with pytest.raises(ValueError):
        sbi_module.seed_sbi_historical(end=tomorrow)


def test_seed_sbi_historical_dry_run(tmp_path: Path) -> None:
    result = sbi_module.seed_sbi_historical(
        db_path=tmp_path / "sbi.db",
        resource_dir=tmp_path,
        dry_run=True,
    )
    assert result.total == 0


def test_seed_sbi_historical_uses_checkpoint(tmp_path: Path, monkeypatch) -> None:
    pdf_path = tmp_path / "latest.pdf"
    pdf_path.write_bytes(b"dummy")

    class _DummyParser:
        def parse(self, path: str | Path):  # type: ignore[override]
            return SBIPDFParseResult(
                rate_date=date(2024, 1, 2),
                rates=[ForexRateRecord(rate_date=date(2024, 1, 2), currency="USD", rate=82.5)],
            )

    class _DummyDownloader:
        def fetch_latest(self) -> Path:  # pragma: no cover - should not be called
            return pdf_path

    class _DummyManager:
        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            return False

        def ingestion_checkpoint(self, _source: str):
            return date(2024, 1, 1)

        def latest_rate_date(self, _source: str):
            return None

        def insert_rates(self, _rows):
            return sbi_module.PersistenceResult(inserted=1)

        def update_ingestion_checkpoint(self, *_args, **_kwargs):
            return None

    monkeypatch.setattr(sbi_module, "SBIPDFParser", _DummyParser)
    monkeypatch.setattr(sbi_module, "SBIPDFDownloader", lambda: _DummyDownloader())
    monkeypatch.setattr(sbi_module, "SQLiteManager", lambda *_args, **_kwargs: _DummyManager())
    monkeypatch.setattr(sbi_module, "_iter_pdf_paths", lambda *_args, **_kwargs: iter([]))

    result = sbi_module.seed_sbi_historical(
        db_path=tmp_path / "sbi.db",
        resource_dir=tmp_path,
        download=True,
    )

    assert result.total == 0


def test_seed_sbi_historical_downloads_latest(tmp_path: Path, monkeypatch) -> None:
    pdf_path = tmp_path / "latest.pdf"
    pdf_path.write_bytes(b"dummy")

    class _DummyParser:
        def parse(self, path: str | Path):  # type: ignore[override]
            return SBIPDFParseResult(
                rate_date=date(2024, 1, 2),
                rates=[ForexRateRecord(rate_date=date(2024, 1, 2), currency="USD", rate=82.5)],
            )

    class _DummyDownloader:
        def fetch_latest(self) -> Path:
            return pdf_path

    class _DummyManager:
        def __enter__(self):
            return self

        def __exit__(self, exc_type, exc, tb):
            return False

        def ingestion_checkpoint(self, _source: str):
            return None

        def latest_rate_date(self, _source: str):
            return None

        def insert_rates(self, _rows):
            return sbi_module.PersistenceResult(inserted=1)

        def update_ingestion_checkpoint(self, *_args, **_kwargs):
            return None

    monkeypatch.setattr(sbi_module, "SBIPDFParser", _DummyParser)
    monkeypatch.setattr(sbi_module, "SBIPDFDownloader", lambda: _DummyDownloader())
    monkeypatch.setattr(sbi_module, "SQLiteManager", lambda *_args, **_kwargs: _DummyManager())
    monkeypatch.setattr(sbi_module, "_iter_pdf_paths", lambda *_args, **_kwargs: iter([]))

    result = sbi_module.seed_sbi_historical(
        db_path=tmp_path / "sbi.db",
        resource_dir=tmp_path,
        download=True,
    )

    assert result.total == 1


def test_seed_sbi_today_persists_pdf(tmp_path: Path, monkeypatch) -> None:
    pdf_path = tmp_path / "latest.pdf"
    pdf_path.write_bytes(b"dummy")
    rates = [ForexRateRecord(rate_date=date(2024, 1, 2), currency="USD", rate=82.5, source="SBI")]

    class _DummyParser:
        def parse(self, path: str | Path):  # type: ignore[override]
            return SBIPDFParseResult(rate_date=date(2024, 1, 2), rates=rates)

    class _DummyDownloader:
        def __init__(self, download_dir: str | Path | None = None) -> None:
            self.download_dir = Path(download_dir) if download_dir else tmp_path

        def fetch_latest(self) -> Path:
            return pdf_path

    monkeypatch.setattr(sbi_module, "SBIPDFParser", _DummyParser)
    monkeypatch.setattr(sbi_module, "SBIPDFDownloader", _DummyDownloader)

    result = sbi_module.seed_sbi_today(db_path=tmp_path / "sbi.db", resource_dir=tmp_path)

    assert result.total == 1
    expected = tmp_path / "2024" / "1" / "2024-01-02.pdf"
    assert expected.exists()


def test_seed_sbi_forex_alias(tmp_path: Path, monkeypatch) -> None:
    monkeypatch.setattr(
        sbi_module,
        "seed_sbi_historical",
        lambda **_kwargs: sbi_module.PersistenceResult(inserted=2),
    )
    result = sbi_module.seed_sbi_forex(db_path=tmp_path / "sbi.db")
    assert result.inserted == 2


def test_seed_sbi_main_invokes_historical(monkeypatch) -> None:
    class _Args:
        db_path = "db"
        resource_dir = "resources"
        start = "2024-01-01"
        end = "2024-01-31"
        download = False

    called = {"value": False}

    def _fake_seed(**_kwargs):
        called["value"] = True
        return sbi_module.PersistenceResult()

    monkeypatch.setattr(sbi_module, "parse_args", lambda: _Args())
    monkeypatch.setattr(sbi_module, "seed_sbi_historical", _fake_seed)

    sbi_module.main()

    assert called["value"] is True



"""PostgreSQL backend strategy."""

from __future__ import annotations

from fx_bharat.db.relational_backend import RelationalBackend


class PostgresBackend(RelationalBackend):
    """Concrete relational backend for PostgreSQL engines."""

    pass


__all__ = ["PostgresBackend"]



"""Parse SBI forex card rate PDFs into ``ForexRateRecord`` rows."""

from __future__ import annotations

import re
import tempfile
import zlib
from dataclasses import dataclass
from datetime import date, datetime
from pathlib import Path
from typing import TYPE_CHECKING, Iterable, cast
from urllib.request import urlretrieve

if TYPE_CHECKING:  # pragma: no cover - type checking only
    from tenacity import RetryError, retry, stop_after_attempt, wait_exponential
else:  # pragma: no cover - optional dependency
    try:
        from tenacity import RetryError, retry, stop_after_attempt, wait_exponential
    except ModuleNotFoundError:

        class RetryError(Exception):
            pass

        def retry(*args, **kwargs):  # type: ignore[no-redef]
            def decorator(func):
                return func

            return decorator

        def stop_after_attempt(*_: int, **__):  # type: ignore[no-redef]
            return None

        def wait_exponential(*_: int, **__):  # type: ignore[no-redef]
            return None


from fx_bharat.ingestion.models import ForexRateRecord
from fx_bharat.utils.logger import get_logger

LOGGER = get_logger(__name__)

SBI_FOREX_PDF_URL = "https://sbi.bank.in/documents/16012/1400784/FOREX_CARD_RATES.pdf"


@dataclass(slots=True)
class SBIPDFParseResult:
    """Represents the parsed content of a single SBI forex PDF."""

    rate_date: date
    rates: list[ForexRateRecord]


class SBIPDFParser:
    """Convert SBI forex PDF documents into structured rows."""

    _CURRENCY_ALIAS_MAP: dict[str, str] = {
        "UAE DIRHAM": "AED",
        "AUS DOLLAR": "AUD",
        "CAD DOLLAR": "CAD",
        "DANISH KRONE": "DKK",
        "EURO": "EUR",
        "HK DOLLAR": "HKD",
        "JAP YEN": "JPY",
        "NOR KRONE": "NOK",
        "NZ DOLLAR": "NZD",
        "SWISS FRANC": "CHF",
        "SG DOLLAR": "SGD",
        "STERLING": "GBP",
        "SA RAND": "ZAR",
        "SAUDI RIYAL": "SAR",
        "SWED KRONA": "SEK",
        "USD": "USD",
    }

    _CURRENCY_PATTERN = re.compile(r"\b([A-Z]{3})\b\s+([0-9]+(?:\.[0-9]+)?)")
    _DATE_PATTERNS = (
        re.compile(r"DATE\s*[:]?\s*(\d{2})[-/](\d{2})[-/](\d{4})"),
        re.compile(r"(\d{4})[-/](\d{2})[-/](\d{2})"),
        re.compile(r"(\d{2})[-/](\d{2})[-/](\d{4})"),
    )

    def parse(self, pdf_path: str | Path, *, rate_date: date | None = None) -> SBIPDFParseResult:
        text = self._extract_text(Path(pdf_path))
        resolved_date = rate_date or self._infer_date(text, pdf_path)
        rows = list(self._extract_rates(text, resolved_date))
        return SBIPDFParseResult(rate_date=resolved_date, rates=rows)

    def _extract_text(self, path: Path) -> str:
        try:
            from pypdf import PdfReader  # type: ignore[import-not-found]
        except ModuleNotFoundError:  # pragma: no cover - dependency missing
            PdfReader = None  # type: ignore[misc, assignment]

        if PdfReader is not None:
            try:
                reader = PdfReader(path)
                return "\n".join(page.extract_text() or "" for page in reader.pages)
            except Exception as exc:  # pragma: no cover - defensive fallback
                LOGGER.warning("Falling back to raw text parsing for %s (%s)", path, exc)

        # Manual PDF stream extraction to cope when ``pypdf`` is unavailable
        raw_bytes = path.read_bytes()
        texts: list[str] = []
        for match in re.finditer(b"stream\r?\n", raw_bytes):
            start = match.end()
            end = raw_bytes.find(b"endstream", start)
            if end == -1:
                continue
            stream = raw_bytes[start:end].lstrip(b"\r\n").rstrip(b"\r\n")
            try:
                stream = zlib.decompress(stream)
            except Exception:
                pass
            for text_match in re.finditer(rb"\(([^)]*)\)", stream):
                try:
                    texts.append(text_match.group(1).decode("latin-1"))
                except Exception:
                    continue

        if texts:
            return "".join(texts)

        try:
            return path.read_text(encoding="utf-8", errors="ignore")
        except Exception as exc:  # pragma: no cover - defensive fallback
            LOGGER.error("Unable to read %s: %s", path, exc)
            raise

    def _infer_date(self, text: str, pdf_path: str | Path) -> date:
        for idx, pattern in enumerate(self._DATE_PATTERNS):
            match = pattern.search(text.upper())
            if not match:
                continue
            groups = match.groups()
            try:
                if idx == 0:
                    day, month, year = (
                        int(groups[0]),
                        int(groups[1]),
                        int(groups[2]),
                    )
                elif idx == 1:
                    year, month, day = (
                        int(groups[0]),
                        int(groups[1]),
                        int(groups[2]),
                    )
                else:
                    day, month, year = (
                        int(groups[0]),
                        int(groups[1]),
                        int(groups[2]),
                    )
                return date(year, month, day)
            except ValueError:
                LOGGER.warning("Ignoring invalid date %s in %s", match.group(0), pdf_path)
                continue
        try:
            return date.fromisoformat(Path(pdf_path).stem)
        except ValueError:
            LOGGER.warning("Falling back to UTC date for %s", pdf_path)
            return datetime.utcnow().date()

    def _extract_rates(self, text: str, rate_date: date) -> Iterable[ForexRateRecord]:
        cleaned_text = re.sub(r"[,\t]+", " ", text.upper())
        cleaned_text = re.sub(r"(\d\.\d{2})(?=\d)", r"\1 ", cleaned_text)
        cleaned_text = re.sub(r"(\d)\s*\.\s*(\d)", r"\1.\2", cleaned_text)
        tokens = sorted(
            set(self._CURRENCY_ALIAS_MAP.keys()) | set(self._CURRENCY_ALIAS_MAP.values()),
            key=len,
            reverse=True,
        )
        token_pattern = r"(" + "|".join(re.escape(token) for token in tokens) + r")"
        cleaned_text = re.sub(token_pattern, r"\n\1", cleaned_text)
        cleaned_text = re.sub(r" +", " ", cleaned_text)
        lines = [line.strip() for line in cleaned_text.splitlines() if line.strip()]

        seen: set[str] = set()

        for line in lines:
            matched = False
            for alias, code in self._CURRENCY_ALIAS_MAP.items():
                if alias not in line:
                    continue
                if code in seen:
                    matched = True
                    break
                numbers = [float(value) for value in re.findall(r"[0-9]+(?:\.[0-9]+)?", line)]
                if len(numbers) < 8:
                    continue
                (
                    tt_buy,
                    tt_sell,
                    bill_buy,
                    bill_sell,
                    travel_card_buy,
                    travel_card_sell,
                    cn_buy,
                    cn_sell,
                ) = numbers[:8]
                yield ForexRateRecord(
                    rate_date=rate_date,
                    currency=code,
                    rate=tt_buy,
                    source="SBI",
                    tt_buy=tt_buy,
                    tt_sell=tt_sell,
                    bill_buy=bill_buy,
                    bill_sell=bill_sell,
                    travel_card_buy=travel_card_buy,
                    travel_card_sell=travel_card_sell,
                    cn_buy=cn_buy,
                    cn_sell=cn_sell,
                )
                seen.add(code)
                matched = True
            if matched:
                continue
            code_match = re.match(r"([A-Z]{3})\b", line)
            if not code_match:
                continue
            code = code_match.group(1)
            if code in seen:
                continue
            numbers = [float(value) for value in re.findall(r"[0-9]+(?:\.[0-9]+)?", line)]
            if len(numbers) < 8:
                continue
            (
                tt_buy,
                tt_sell,
                bill_buy,
                bill_sell,
                travel_card_buy,
                travel_card_sell,
                cn_buy,
                cn_sell,
            ) = numbers[:8]
            yield ForexRateRecord(
                rate_date=rate_date,
                currency=code,
                rate=tt_buy,
                source="SBI",
                tt_buy=tt_buy,
                tt_sell=tt_sell,
                bill_buy=bill_buy,
                bill_sell=bill_sell,
                travel_card_buy=travel_card_buy,
                travel_card_sell=travel_card_sell,
                cn_buy=cn_buy,
                cn_sell=cn_sell,
            )
            seen.add(code)


class SBIPDFDownloader:
    """Download the latest SBI forex PDF from the public endpoint."""

    def __init__(self, download_dir: str | Path | None = None) -> None:
        self.download_dir = Path(download_dir) if download_dir else Path(tempfile.mkdtemp())
        self.download_dir.mkdir(parents=True, exist_ok=True)

    def fetch_latest(self, destination: str | Path | None = None) -> Path:
        """Fetch the PDF and optionally persist it to ``destination``.

        When ``destination`` is omitted, the PDF is stored in ``download_dir``.
        """

        destination_path = (
            Path(destination) if destination else self.download_dir / Path(SBI_FOREX_PDF_URL).name
        )
        destination_path.parent.mkdir(parents=True, exist_ok=True)
        LOGGER.info("Downloading SBI forex PDF to %s", destination_path)
        try:
            return cast(Path, self._download_with_retry(destination_path))
        except RetryError as exc:  # pragma: no cover - network variability
            LOGGER.error("Exhausted retries while downloading SBI PDF: %s", exc)
            last_exception = exc.last_attempt.exception()
            if last_exception is not None:
                raise last_exception
            raise RuntimeError("RetryError raised without underlying exception")

    @retry(wait=wait_exponential(multiplier=1, min=2, max=30), stop=stop_after_attempt(3))
    def _download_with_retry(self, destination_path: Path) -> Path:
        urlretrieve(SBI_FOREX_PDF_URL, destination_path)
        return destination_path


__all__ = ["SBI_FOREX_PDF_URL", "SBIPDFDownloader", "SBIPDFParser", "SBIPDFParseResult"]



from __future__ import annotations

from datetime import date

import pytest

from fx_bharat import DatabaseBackend, DatabaseConnectionInfo, FxBharat
from fx_bharat.ingestion.models import ForexRateRecord


@pytest.fixture()
def sqlite_fx(tmp_path):
    db_path = tmp_path / "fx.db"
    config = DatabaseConnectionInfo(
        backend=DatabaseBackend.SQLITE,
        url=f"sqlite:///{db_path}",
        name=str(db_path),
        username=None,
        password=None,
        host=None,
        port=None,
    )
    return FxBharat(db_config=config)


def _seed_sample_data(app: FxBharat) -> None:
    assert app.sqlite_manager is not None
    app.sqlite_manager.insert_rates(
        [
            ForexRateRecord(rate_date=date(2024, 1, 1), currency="USD", rate=82.0, source="RBI"),
            ForexRateRecord(rate_date=date(2024, 1, 1), currency="USD", rate=81.5, source="SBI"),
            ForexRateRecord(rate_date=date(2024, 1, 2), currency="USD", rate=82.2, source="RBI"),
        ]
    )


def test_history_source_filter(sqlite_fx: FxBharat) -> None:
    _seed_sample_data(sqlite_fx)

    snapshots = sqlite_fx.history(date(2024, 1, 1), date(2024, 1, 2), source_filter="sbi")

    assert all(snap["source"] == "SBI" for snap in snapshots)
    assert len(snapshots) == 1


def test_rate_sorts_and_blends(sqlite_fx: FxBharat) -> None:
    _seed_sample_data(sqlite_fx)

    snapshots = sqlite_fx.rate()

    assert snapshots[0]["source"] == "SBI"
    assert snapshots[1]["source"] == "RBI"
    assert snapshots[0]["rate_date"] <= snapshots[1]["rate_date"]



"""CLI + helpers for populating (seeding) the forex database from RBI archives."""

from __future__ import annotations

import argparse
from datetime import timedelta
from pathlib import Path

from fx_bharat.db import DEFAULT_SQLITE_DB_PATH
from fx_bharat.db.sqlite_manager import PersistenceResult, SQLiteManager
from fx_bharat.ingestion.rbi_csv import RBICSVParser
from fx_bharat.ingestion.rbi_selenium import RBINoReferenceRateError, RBISeleniumClient
from fx_bharat.ingestion.rbi_workbook import RBIWorkbookConverter
from fx_bharat.utils.date_range import month_ranges, parse_date
from fx_bharat.utils.logger import get_logger
from fx_bharat.utils.rbi import enforce_rbi_min_date

LOGGER = get_logger(__name__)

__all__ = ["seed_rbi_forex", "parse_args", "main"]


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--from", dest="start", required=True, help="Start date (YYYY-MM-DD)")
    parser.add_argument("--to", dest="end", required=True, help="End date (YYYY-MM-DD)")
    parser.add_argument(
        "--db",
        dest="db_path",
        default=str(DEFAULT_SQLITE_DB_PATH),
        help="SQLite database path",
    )
    parser.add_argument(
        "--no-headless",
        dest="headless",
        action="store_false",
        default=True,
        help="Disable headless Chrome mode",
    )
    parser.add_argument("--download-dir", dest="download_dir", help="Optional download directory")
    return parser.parse_args()


def _log_chunk_result(chunk_label: str, result: PersistenceResult) -> None:
    LOGGER.info(
        "%s → inserted %s rows, updated %s rows (total %s)",
        chunk_label,
        result.inserted,
        result.updated,
        result.total,
    )


def seed_rbi_forex(
    start: str,
    end: str,
    *,
    db_path: str | Path = DEFAULT_SQLITE_DB_PATH,
    headless: bool = True,
    download_dir: str | Path | None = None,
    incremental: bool = True,
    dry_run: bool = False,
) -> PersistenceResult:
    """Seed RBI forex data between ``start`` and ``end`` dates."""

    start_date = parse_date(start)
    end_date = parse_date(end)
    enforce_rbi_min_date(start_date, end_date)
    if dry_run:
        LOGGER.info("Dry-run enabled; skipping RBI ingestion for %s → %s", start, end)
        return PersistenceResult()
    converter = RBIWorkbookConverter()
    csv_parser = RBICSVParser()
    download_path = Path(download_dir) if download_dir else None
    total = PersistenceResult()
    with SQLiteManager(db_path) as manager:
        effective_start = start_date
        if incremental:
            checkpoint = manager.ingestion_checkpoint("RBI") or manager.latest_rate_date("RBI")
            if checkpoint and checkpoint >= start_date:
                effective_start = checkpoint + timedelta(days=1)
        if effective_start > end_date:
            LOGGER.info("RBI data already ingested up to %s; nothing to do", end_date)
            return total
        date_chunks = list(month_ranges(effective_start, end_date))
        with RBISeleniumClient(download_dir=download_path, headless=headless) as client:
            for chunk in date_chunks:
                LOGGER.info("Processing %s - %s", chunk.start, chunk.end)
                try:
                    excel_path = client.fetch_excel(chunk.start, chunk.end)
                except RBINoReferenceRateError as exc:
                    LOGGER.warning(
                        "RBI reference rates for %s → %s are not yet published; stopping ingestion early (%s)",
                        chunk.start,
                        chunk.end,
                        exc,
                    )
                    break
                csv_path = converter.to_csv(
                    excel_path,
                    start_date=chunk.start,
                    end_date=chunk.end,
                    output_dir=client.download_dir,
                )
                csv_rows = csv_parser.parse(csv_path)
                result = manager.insert_rates(csv_rows)
                _log_chunk_result(f"Chunk {chunk.start} → {chunk.end}", result)
                total.inserted += result.inserted
                total.updated += result.updated
                if result.inserted:
                    latest_day = max(row.rate_date for row in csv_rows)
                    manager.update_ingestion_checkpoint("RBI", latest_day)
    LOGGER.info(
        "Seeding finished: inserted %s rows, updated %s rows (total %s)",
        total.inserted,
        total.updated,
        total.total,
    )
    return total


def main() -> None:
    args = parse_args()
    seed_rbi_forex(
        args.start,
        args.end,
        db_path=args.db_path,
        headless=args.headless,
        download_dir=args.download_dir,
    )


if __name__ == "__main__":  # pragma: no cover - CLI entry point
    main()



from __future__ import annotations

import sqlite3
from datetime import date

import pytest

from fx_bharat.db import sqlite_manager as sqlite_manager_module
from fx_bharat.db.sqlite_manager import SQLiteManager
from fx_bharat.ingestion.models import ForexRateRecord, LmeRateRecord


def test_sqlalchemy_backend_patches_lme_schema(tmp_path) -> None:
    db_path = tmp_path / "legacy.db"
    with sqlite3.connect(db_path) as conn:
        conn.executescript(
            """
            CREATE TABLE lme_copper_rates (rate_date DATE PRIMARY KEY);
            CREATE TABLE lme_aluminum_rates (rate_date DATE PRIMARY KEY);
            """
        )

    manager = SQLiteManager(db_path)
    if not isinstance(manager._backend, sqlite_manager_module._SQLAlchemyBackend):
        manager.close()
        pytest.skip("SQLAlchemy backend not active")
    with sqlite3.connect(db_path) as conn:
        columns = {
            row[1] for row in conn.execute("PRAGMA table_info(lme_aluminum_rates)").fetchall()
        }
    manager.close()

    assert {"price", "price_3_month", "stock", "created_at"}.issubset(columns)


def test_sqlalchemy_backend_updates_sbi_rows(tmp_path) -> None:
    manager = SQLiteManager(tmp_path / "updates.db")
    try:
        row = ForexRateRecord(
            rate_date=date(2024, 1, 1),
            currency="USD",
            rate=80.0,
            source="SBI",
            tt_buy=1.0,
            tt_sell=2.0,
        )
        manager.insert_rates([row])
        updated = ForexRateRecord(
            rate_date=date(2024, 1, 1),
            currency="USD",
            rate=81.0,
            source="SBI",
            tt_buy=2.0,
            tt_sell=3.0,
        )
        result = manager.insert_rates([updated])
    finally:
        manager.close()

    assert result.total == 1


def test_fallback_backend_lme_and_metadata(tmp_path, monkeypatch) -> None:
    monkeypatch.setattr(sqlite_manager_module, "SQLALCHEMY_AVAILABLE", False)
    manager = sqlite_manager_module.SQLiteManager(tmp_path / "fallback_lme.db")
    try:
        rows = [
            LmeRateRecord(
                rate_date=date(2024, 2, 1),
                price=100.0,
                price_3_month=101.0,
                stock=10,
                metal="COPPER",
            )
        ]
        manager.insert_lme_rates("COPPER", rows)
        updated = [
            LmeRateRecord(
                rate_date=date(2024, 2, 1),
                price=200.0,
                price_3_month=201.0,
                stock=20,
                metal="COPPER",
            )
        ]
        result = manager.insert_lme_rates("COPPER", updated)
        fetched = manager.fetch_lme_range("COPPER", start=date(2024, 2, 1), end=date(2024, 2, 1))
        latest = manager.latest_rate_date("SBI")
        manager.update_ingestion_checkpoint("LME_COPPER", date(2024, 2, 1))
        checkpoint = manager.ingestion_checkpoint("LME_COPPER")
    finally:
        manager.close()

    assert result.total == 1
    assert fetched[0].price == 200.0
    assert latest is None
    assert checkpoint == date(2024, 2, 1)



"""CSV helpers for storing and re-reading RBI downloads."""

from __future__ import annotations

import csv
from collections import defaultdict
from datetime import date, datetime
from pathlib import Path
from typing import Iterable, Sequence, cast

from fx_bharat.ingestion.models import ForexRateRecord

EXPECTED_CURRENCIES = ("USD", "GBP", "EUR", "JPY")
CSV_HEADER = ("Date", "USD", "GBP", "EURO", "YEN")
ISO_TO_CSV = {
    "USD": "USD",
    "GBP": "GBP",
    "EUR": "EURO",
    "JPY": "YEN",
}
CSV_TO_ISO = {value: key for key, value in ISO_TO_CSV.items()}


class RBICSVExporter:
    """Export normalized forex rows into a RBI-style CSV file."""

    def __init__(self, *, date_format: str = "%d/%m/%Y") -> None:
        self.date_format = date_format

    def write(
        self,
        records: Sequence[ForexRateRecord],
        *,
        start_date: date,
        end_date: date,
        output_dir: Path | None = None,
    ) -> Path:
        if not records:
            raise ValueError("records collection is empty")

        directory = Path(output_dir) if output_dir else Path.cwd()
        directory.mkdir(parents=True, exist_ok=True)
        csv_name = self._build_filename(start_date, end_date)
        csv_path = directory / csv_name
        rows_by_date: dict[date, dict[str, float]] = defaultdict(dict)
        for record in records:
            rows_by_date[record.rate_date][record.currency] = record.rate

        with csv_path.open("w", newline="", encoding="utf-8") as handle:
            writer = csv.writer(handle)
            writer.writerow(CSV_HEADER)
            for rate_date in sorted(rows_by_date):
                date_str = rate_date.strftime(self.date_format)
                row = [date_str]
                for currency in EXPECTED_CURRENCIES:
                    value = rows_by_date[rate_date].get(currency)
                    row.append("" if value is None else f"{value}")
                writer.writerow(row)
        return csv_path

    def _build_filename(self, start_date: date, end_date: date) -> str:
        safe_from = start_date.strftime("%d-%m-%Y")
        safe_to = end_date.strftime("%d-%m-%Y")
        return f"RBI_Reference_Rates_{safe_from}_to_{safe_to}.csv"


class RBICSVParser:
    """Parse CSV files produced by :class:`RBICSVExporter`."""

    def __init__(self, *, date_format: str = "%d/%m/%Y") -> None:
        self.date_format = date_format

    def parse(self, csv_path: str | Path) -> list[ForexRateRecord]:
        path = Path(csv_path)
        if not path.exists():
            raise FileNotFoundError(path)

        with path.open("r", newline="", encoding="utf-8") as handle:
            reader = csv.DictReader(handle)
            self._validate_header(reader.fieldnames)
            rows: list[ForexRateRecord] = []
            for row in reader:
                date_raw = row.get("Date")
                if not date_raw:
                    continue
                rate_date = datetime.strptime(date_raw.strip(), self.date_format).date()
                for csv_column, iso_code in CSV_TO_ISO.items():
                    value_raw = row.get(csv_column)
                    if value_raw in (None, ""):
                        continue
                    try:
                        rate = float(cast(str, value_raw))
                    except ValueError:
                        continue
                    rows.append(ForexRateRecord(rate_date=rate_date, currency=iso_code, rate=rate))
        return rows

    @staticmethod
    def _validate_header(fieldnames: Iterable[str] | None) -> None:
        if not fieldnames:
            raise ValueError("CSV file does not contain a header row")
        normalized = [field.strip() for field in fieldnames]
        if normalized[: len(CSV_HEADER)] != list(CSV_HEADER):
            raise ValueError("Unexpected CSV header format")



"""Utility helpers for generating RBI-friendly date ranges."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import date, datetime, timedelta
from typing import Iterator, Tuple


@dataclass(frozen=True)
class DateRange:
    """Container representing a closed date range."""

    start: date
    end: date

    def as_tuple(self) -> Tuple[date, date]:
        """Return the range as a tuple of ``(start, end)``."""
        return (self.start, self.end)


def parse_date(value: str | date) -> date:
    """Parse a date string in ISO format to :class:`date`."""

    if isinstance(value, date):
        return value
    return datetime.strptime(value, "%Y-%m-%d").date()


def month_ranges(start: str | date, end: str | date) -> Iterator[DateRange]:
    """Yield date ranges aligned by month within the inclusive window."""

    start_date = parse_date(start)
    end_date = parse_date(end)
    if start_date > end_date:
        raise ValueError("start date must not be after end date")

    current = start_date
    while current <= end_date:
        month_end = _end_of_month(current)
        chunk_end = min(month_end, end_date)
        yield DateRange(start=current, end=chunk_end)
        current = chunk_end + timedelta(days=1)


def split_ranges(start: str | date, end: str | date, window_days: int) -> Iterator[DateRange]:
    """Split the period into smaller windows of ``window_days`` length."""

    if window_days <= 0:
        raise ValueError("window_days must be positive")

    start_date = parse_date(start)
    end_date = parse_date(end)
    if start_date > end_date:
        raise ValueError("start date must not be after end date")

    current = start_date
    delta = timedelta(days=window_days - 1)
    while current <= end_date:
        chunk_end = min(current + delta, end_date)
        yield DateRange(start=current, end=chunk_end)
        current = chunk_end + timedelta(days=1)


def _end_of_month(day: date) -> date:
    """Return the last day of the month for ``day``."""

    if day.month == 12:
        return date(day.year, 12, 31)
    first_next_month = date(day.year, day.month + 1, 1)
    return first_next_month - timedelta(days=1)



from __future__ import annotations

from fx_bharat.db.relational_backend import RelationalBackend


class _DummyResult:
    def __init__(self, rows=None, scalar_value=None) -> None:
        self._rows = rows or []
        self._scalar_value = scalar_value

    def __iter__(self):
        return iter(self._rows)

    def scalar(self):
        return self._scalar_value


class _DummyDialect:
    def __init__(self, name: str) -> None:
        self.name = name


class _DummyEngine:
    def __init__(self, name: str) -> None:
        self.dialect = _DummyDialect(name)


class _DummyConnection:
    def __init__(self, dialect_name: str, existing: set[str]) -> None:
        self.engine = _DummyEngine(dialect_name)
        self.existing = existing
        self.executed: list[str] = []

    def execute(self, stmt, params=None):
        sql = str(stmt)
        self.executed.append(sql)
        if "SELECT DATABASE()" in sql:
            return _DummyResult(scalar_value="testdb")
        if "information_schema.columns" in sql:
            return _DummyResult(rows=[(col,) for col in self.existing])
        return _DummyResult()


def test_relational_backend_schema_patch_postgres() -> None:
    backend = RelationalBackend("sqlite:///:memory:")
    connection = _DummyConnection("postgresql", {"rate_date", "usd_price"})

    backend._ensure_lme_schema(connection)

    assert any(
        "ALTER TABLE lme_copper_rates ADD COLUMN price" in sql for sql in connection.executed
    )
    assert any(
        "ALTER TABLE lme_copper_rates DROP COLUMN IF EXISTS usd_price" in sql
        for sql in connection.executed
    )


def test_relational_backend_schema_patch_mysql() -> None:
    backend = RelationalBackend("sqlite:///:memory:")
    connection = _DummyConnection("mysql", {"rate_date", "usd_price"})

    backend._ensure_lme_schema(connection)

    assert any(
        "ALTER TABLE lme_aluminum_rates ADD COLUMN price" in sql for sql in connection.executed
    )
    assert any(
        "ALTER TABLE lme_aluminum_rates DROP COLUMN usd_price" in sql for sql in connection.executed
    )



"""Ingestion helpers for London Metal Exchange cash seller prices."""

from __future__ import annotations

import re
from dataclasses import dataclass
from datetime import date
from typing import Iterable, Literal, cast

import pandas as pd
import requests  # type: ignore[import-untyped]
from bs4 import BeautifulSoup
from bs4.element import Tag

from fx_bharat.ingestion.models import LmeRateRecord
from fx_bharat.utils.logger import get_logger

LOGGER = get_logger(__name__)

LME_URLS: dict[Literal["COPPER", "ALUMINUM"], str] = {
    "COPPER": "https://www.westmetall.com/en/markdaten.php?action=table&field=LME_Cu_cash",
    "ALUMINUM": "https://www.westmetall.com/en/markdaten.php?action=table&field=LME_Al_cash",
}


@dataclass(slots=True)
class LmeTableParseResult:
    metal: Literal["COPPER", "ALUMINUM"]
    rows: list[LmeRateRecord]


def _normalise_metal(metal: str) -> Literal["COPPER", "ALUMINUM"]:
    upper = metal.upper()
    if upper in {"CU", "COPPER"}:
        return "COPPER"
    if upper in {"AL", "ALUMINUM", "ALUMINIUM"}:
        return "ALUMINUM"
    raise ValueError(f"Unsupported LME metal: {metal}")


def _parse_float(value: object | None) -> float | None:
    if value is None:
        return None
    if isinstance(value, (int, float)):
        try:
            return float(value)
        except (TypeError, ValueError):
            return None
    cleaned = re.sub(r"[^0-9,.-]", "", str(value))
    cleaned = cleaned.replace(",", "")
    try:
        return float(cleaned)
    except (TypeError, ValueError):
        return None


def _parse_int(value: object | None) -> int | None:
    parsed = _parse_float(value)
    if parsed is None:
        return None
    try:
        return int(parsed)
    except (TypeError, ValueError):
        return None


def _coerce_date(value: object) -> date | None:
    if isinstance(value, str):
        value = re.sub(r"\s+", " ", value.strip())
        value = re.sub(r"([A-Za-z]{2,})\s+([A-Za-z]{2,})", r"\1\2", value)
    try:
        parsed = pd.to_datetime(value, errors="coerce", dayfirst=True)
    except (TypeError, ValueError):
        return None
    if pd.isna(parsed):
        return None
    return cast(date, parsed.date())


def _find_column(columns: Iterable[str], keywords: set[str]) -> str | None:
    for column in columns:
        lower = column.lower()
        if any(keyword in lower for keyword in keywords):
            return column
    return None


def parse_lme_table(html: str, metal: str) -> LmeTableParseResult:
    """Parse Westmetall HTML tables into structured records.

    The source pages render one table per year (anchored by ``#yYYYY``) and
    sometimes repeat header rows inside each table body. Rather than relying on
    :func:`pandas.read_html` to guess the header structure, we parse tables with
    BeautifulSoup to preserve every data row and then map the appropriate
    numeric columns (cash-settlement first, otherwise any numeric column).
    """

    def _cell_text(cell: Tag) -> str:
        return " ".join(cell.stripped_strings).strip()

    normalised = _normalise_metal(metal)
    soup = BeautifulSoup(html, "html.parser")

    tables: list[Tag] = []
    for anchor in soup.find_all("a", id=re.compile(r"^y\d{4}$")):
        table = anchor.find_next("table")
        if table:
            tables.append(table)
    if not tables:
        tables = list(soup.find_all("table"))
    if not tables:
        raise ValueError("No tables found in supplied HTML")

    dataframes: list[pd.DataFrame] = []
    for table in tables:
        headers = [_cell_text(th) for th in table.find_all("th") if _cell_text(th)]
        table_rows: list[list[str]] = []
        for tr in table.find_all("tr"):
            cols = tr.find_all("td")
            if not cols:
                continue
            table_rows.append([_cell_text(col) for col in cols])
        if not table_rows:
            continue
        if headers and len(headers) == len(table_rows[0]):
            frame = pd.DataFrame(table_rows, columns=headers)
        else:
            frame = pd.DataFrame(table_rows)
        dataframes.append(frame)

    if not dataframes:
        raise ValueError("No data rows found in supplied HTML")

    def _numeric_columns(frame: pd.DataFrame, *, exclude: set[str]) -> list[str]:
        numerics: list[str] = []
        for column in frame.columns:
            if column in exclude:
                continue
            series = frame[column]
            if series.apply(_parse_float).notna().any():
                numerics.append(str(column))
        return numerics

    rows: list[LmeRateRecord] = []
    for frame in dataframes:
        frame.columns = [str(col).strip() for col in frame.columns]
        date_col = _find_column(frame.columns, {"date", "datum"}) or frame.columns[0]
        price_col = _find_column(frame.columns, {"cash", "settlement", "seller"})
        three_month_col = _find_column(frame.columns, {"3-month", "3 month", "3m"})
        stock_col = _find_column(frame.columns, {"stock", "stocks"})

        numeric_cols = _numeric_columns(frame, exclude={date_col})
        if price_col is None and numeric_cols:
            price_col = numeric_cols[0]
        if three_month_col is None and len(numeric_cols) > 1:
            fallback_index = 1 if numeric_cols[0] == price_col else 0
            three_month_candidates = [col for col in numeric_cols if col != price_col]
            if three_month_candidates:
                three_month_col = three_month_candidates[0]
            elif len(numeric_cols) > fallback_index:
                three_month_col = numeric_cols[fallback_index]
        if stock_col is None and numeric_cols:
            candidates = [col for col in numeric_cols if col not in {price_col, three_month_col}]
            if candidates:
                stock_col = candidates[-1]
        for _, row in frame.iterrows():
            rate_date = _coerce_date(row.get(date_col))
            if rate_date is None:
                continue
            record = LmeRateRecord(
                rate_date=rate_date,
                price=_parse_float(row.get(price_col)) if price_col else None,
                price_3_month=_parse_float(row.get(three_month_col)) if three_month_col else None,
                stock=_parse_int(row.get(stock_col)) if stock_col else None,
                metal=normalised,
            )
            rows.append(record)
    return LmeTableParseResult(metal=normalised, rows=rows)


def fetch_lme_rates(metal: str, *, session: requests.Session | None = None) -> LmeTableParseResult:
    """Download LME cash seller table and return parsed rows."""

    normalised = _normalise_metal(metal)
    url = LME_URLS[normalised]
    sess = session or requests.Session()
    sess.headers.setdefault("User-Agent", "fx-bharat-lme-ingestor/1.0")
    response = sess.get(url, timeout=30)
    response.raise_for_status()
    LOGGER.info("Fetched %s data from %s", normalised, url)
    return parse_lme_table(response.text, normalised)


__all__ = ["fetch_lme_rates", "parse_lme_table", "LmeTableParseResult", "LME_URLS"]



"""RBI-specific helpers and invariants used across the package."""

from __future__ import annotations

from datetime import date

RBI_MIN_AVAILABLE_DATE = date(2022, 4, 12)
RBI_MIN_DATE_MESSAGE = "RBI do not provide the data before 12/04/2022."


def enforce_rbi_min_date(*dates: date) -> None:
    """Ensure all provided dates are on/after :data:`RBI_MIN_AVAILABLE_DATE`."""

    if any(day < RBI_MIN_AVAILABLE_DATE for day in dates):
        raise ValueError(RBI_MIN_DATE_MESSAGE)


__all__ = ["RBI_MIN_AVAILABLE_DATE", "RBI_MIN_DATE_MESSAGE", "enforce_rbi_min_date"]



"""Mongo backend tests that monkeypatch pymongo primitives."""

from __future__ import annotations

from datetime import date
from pathlib import Path
from typing import Any, Dict, List

import pytest

from fx_bharat.db import mongo_backend as mongo_module
from fx_bharat.ingestion.models import ForexRateRecord, LmeRateRecord


class _DummyCursor:
    def __init__(self, docs: List[Dict[str, Any]]) -> None:
        self._docs = docs

    def sort(self, field: str, direction: int) -> List[Dict[str, Any]]:
        reverse = direction == -1
        return sorted(self._docs, key=lambda doc: doc[field], reverse=reverse)


class _DummyCollection:
    def __init__(self, name: str = "") -> None:
        self.docs: Dict[tuple[str, str] | str, Dict[str, Any]] = {}
        self.indexes: list[tuple[tuple[tuple[str, int], ...], bool]] = []
        self.name = name

    def __bool__(self) -> bool:  # pragma: no cover - behavioural parity with pymongo
        raise NotImplementedError("Collection truthiness is undefined")

    def create_index(self, fields: list[tuple[str, int]], unique: bool) -> None:
        self.indexes.append((tuple(fields), unique))

    def bulk_write(self, operations: list["_DummyUpdateOne"], ordered: bool) -> "_DummyBulkResult":
        assert ordered is False
        for op in operations:
            assert isinstance(op, _DummyUpdateOne)
            if "currency_code" in op.filter:
                key = (op.filter["rate_date"], op.filter["currency_code"])
            else:
                key = op.filter["rate_date"]
            self.docs[key] = dict(op.update["$set"])
        return _DummyBulkResult()

    def find(self, query: Dict[str, Dict[str, str]]):  # type: ignore[override]
        docs = list(self.docs.values())
        if "rate_date" in query:
            range_query = query["rate_date"]
            if "$gte" in range_query:
                docs = [doc for doc in docs if doc["rate_date"] >= range_query["$gte"]]
            if "$lte" in range_query:
                docs = [doc for doc in docs if doc["rate_date"] <= range_query["$lte"]]
        return _DummyCursor(docs)


class _DummyBulkResult:
    inserted_count = 0
    upserted_count = 0
    modified_count = 0


class _DummyUpdateOne:
    def __init__(
        self, filter: Dict[str, str], update: Dict[str, Dict[str, Any]], *, upsert: bool
    ) -> None:
        assert upsert is True
        self.filter = filter
        self.update = update
        self.upsert = upsert


class _DummyDatabase(dict):
    def __getitem__(self, name: str) -> _DummyCollection:  # type: ignore[override]
        if name not in self:
            self[name] = _DummyCollection(name=name)
        return dict.__getitem__(self, name)


class _DummyClient:
    def __init__(self, url: str) -> None:
        self.url = url
        self.admin = self
        self.closed = False
        self.databases: Dict[str, _DummyDatabase] = {}

    def __getitem__(self, name: str) -> _DummyDatabase:
        return self.databases.setdefault(name, _DummyDatabase())

    def get_default_database(self) -> _DummyDatabase:
        return self.__getitem__("default")

    def command(self, name: str) -> None:
        assert name == "ping"

    def close(self) -> None:
        self.closed = True


@pytest.fixture(autouse=True)
def patch_mongo_client(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.setattr(mongo_module, "MongoClient", _DummyClient)
    monkeypatch.setattr(mongo_module, "PyMongoError", RuntimeError)
    monkeypatch.setattr(mongo_module, "UpdateOne", _DummyUpdateOne)


def test_mongo_backend_roundtrip(tmp_path: Path) -> None:
    backend = mongo_module.MongoBackend("mongodb://example.com/", database="fx")
    backend.ensure_schema()

    rows = [
        ForexRateRecord(rate_date=date(2024, 1, 1), currency="USD", rate=82.5),
        ForexRateRecord(rate_date=date(2024, 1, 2), currency="EUR", rate=89.1),
    ]
    result = backend.insert_rates(rows)
    assert result.inserted == 2

    fetched = backend.fetch_range(date(2024, 1, 1), date(2024, 1, 2))
    assert len(fetched) == 2
    assert fetched[0].currency == "USD"
    assert fetched[1].currency == "EUR"

    backend.close()


def test_mongo_backend_lme_roundtrip() -> None:
    backend = mongo_module.MongoBackend("mongodb://example.com/", database="fx")
    backend.ensure_schema()

    rows = [
        LmeRateRecord(
            rate_date=date(2024, 2, 1),
            price=8500.0,
            price_3_month=8450.0,
            stock=120,
            metal="COPPER",
        )
    ]
    result = backend.insert_lme_rates("CU", rows)
    assert result.inserted == 1

    fetched = backend.fetch_lme_range("COPPER")
    assert len(fetched) == 1
    assert fetched[0].price == 8500.0
    assert fetched[0].metal == "COPPER"

    backend.close()


def test_mongo_backend_handles_empty_inputs() -> None:
    backend = mongo_module.MongoBackend("mongodb://example.com/", database="fx")
    result = backend.insert_rates([])
    assert result.total == 0
    backend.close()


def test_mongo_backend_fetch_range_uses_collection_name() -> None:
    backend = mongo_module.MongoBackend("mongodb://example.com/", database="fx")
    backend._collection = _DummyCollection(name="forex_rates_sbi")
    backend._sbi_collection = None
    backend._rbi_collection = None
    backend._collection.docs[("2024-01-01", "USD")] = {
        "rate_date": "2024-01-01",
        "currency_code": "USD",
        "rate": 82.5,
    }

    fetched = backend.fetch_range(source="SBI")
    assert fetched[0].source == "SBI"
    backend.close()


def test_mongo_backend_fetch_lme_range_with_dates() -> None:
    backend = mongo_module.MongoBackend("mongodb://example.com/", database="fx")
    backend._lme_copper_collection = _DummyCollection(name="lme_copper_rates")
    backend._lme_copper_collection.docs["2024-01-01"] = {
        "rate_date": "2024-01-01",
        "price": 8500.0,
        "price_3_month": 8450.0,
        "stock": 100,
    }
    rows = backend.fetch_lme_range("COPPER", start=date(2024, 1, 1), end=date(2024, 1, 1))
    assert rows[0].price == 8500.0
    backend.close()



from __future__ import annotations

from datetime import date

from fx_bharat.seeds import populate_rbi_forex as rbi_module


def test_seed_rbi_forex_dry_run(tmp_path) -> None:
    result = rbi_module.seed_rbi_forex(
        "2024-01-01",
        "2024-01-02",
        db_path=tmp_path / "rbi.db",
        dry_run=True,
    )
    assert result.total == 0


def test_seed_rbi_forex_skips_when_checkpoint_ahead(tmp_path) -> None:
    db_path = tmp_path / "rbi.db"
    with rbi_module.SQLiteManager(db_path) as manager:
        manager.update_ingestion_checkpoint("RBI", date(2024, 1, 10))

    result = rbi_module.seed_rbi_forex(
        "2024-01-01",
        "2024-01-02",
        db_path=db_path,
        incremental=True,
    )

    assert result.total == 0


def test_rbi_parse_args_defaults(monkeypatch) -> None:
    monkeypatch.setattr(
        "sys.argv",
        [
            "prog",
            "--from",
            "2024-01-01",
            "--to",
            "2024-01-31",
        ],
    )
    args = rbi_module.parse_args()
    assert args.headless is True
    assert args.download_dir is None



from __future__ import annotations

from datetime import date

import pytest

from fx_bharat.db.base_backend import BackendStrategy
from fx_bharat.db.sqlite_manager import PersistenceResult
from fx_bharat.ingestion.models import ForexRateRecord


class _DummyBackend(BackendStrategy):
    def ensure_schema(self) -> None:
        return None

    def insert_rates(self, rows: list[ForexRateRecord]) -> PersistenceResult:
        return PersistenceResult(inserted=len(rows))

    def fetch_range(
        self,
        start: date | None = None,
        end: date | None = None,
        *,
        source: str | None = None,
    ) -> list[ForexRateRecord]:
        return []


def test_base_backend_lme_methods_raise_not_implemented() -> None:
    backend = _DummyBackend()

    with pytest.raises(NotImplementedError):
        backend.insert_lme_rates("COPPER", [])
    with pytest.raises(NotImplementedError):
        backend.fetch_lme_range("COPPER")



"""Persistence helpers for fx_bharat (prefers SQLAlchemy)."""

from __future__ import annotations

import sqlite3
from dataclasses import dataclass
from datetime import date
from pathlib import Path
from typing import TYPE_CHECKING, Any, Protocol, Sequence, cast

from fx_bharat.db import DEFAULT_SQLITE_DB_PATH
from fx_bharat.ingestion.models import ForexRateRecord, LmeRateRecord
from fx_bharat.utils.logger import get_logger

LOGGER = get_logger(__name__)

try:  # pragma: no cover - exercised indirectly
    from sqlalchemy import Column, Date, DateTime, Float, String, create_engine, select, text
    from sqlalchemy.orm import DeclarativeBase, Session, sessionmaker
    from sqlalchemy.sql import func
except ModuleNotFoundError:  # pragma: no cover - fallback path
    SQLALCHEMY_AVAILABLE = False
else:  # pragma: no cover - module import time
    SQLALCHEMY_AVAILABLE = True

    class Base(DeclarativeBase):
        pass

    class _RbiRate(Base):
        __tablename__ = "forex_rates_rbi"

        rate_date = Column(Date, primary_key=True)
        currency = Column(String, primary_key=True)
        rate = Column(Float, nullable=False)
        created_at = Column(DateTime, server_default=text("CURRENT_TIMESTAMP"))

    class _SbiRate(Base):
        __tablename__ = "forex_rates_sbi"

        rate_date = Column(Date, primary_key=True)
        currency = Column(String, primary_key=True)
        rate = Column(Float, nullable=False)
        tt_buy = Column(Float, nullable=True)
        tt_sell = Column(Float, nullable=True)
        bill_buy = Column(Float, nullable=True)
        bill_sell = Column(Float, nullable=True)
        travel_card_buy = Column(Float, nullable=True)
        travel_card_sell = Column(Float, nullable=True)
        cn_buy = Column(Float, nullable=True)
        cn_sell = Column(Float, nullable=True)
        created_at = Column(DateTime, server_default=text("CURRENT_TIMESTAMP"))

    class _IngestionMetadata(Base):
        __tablename__ = "ingestion_metadata"

        source = Column(String, primary_key=True)
        last_ingested_date = Column(Date, nullable=False)
        updated_at = Column(DateTime, server_default=text("CURRENT_TIMESTAMP"))

    class _LmeCopperRate(Base):
        __tablename__ = "lme_copper_rates"

        rate_date = Column(Date, primary_key=True)
        price = Column(Float, nullable=True)
        price_3_month = Column(Float, nullable=True)
        stock = Column(Float, nullable=True)
        created_at = Column(DateTime, server_default=text("CURRENT_TIMESTAMP"))

    class _LmeAluminumRate(Base):
        __tablename__ = "lme_aluminum_rates"

        rate_date = Column(Date, primary_key=True)
        price = Column(Float, nullable=True)
        price_3_month = Column(Float, nullable=True)
        stock = Column(Float, nullable=True)
        created_at = Column(DateTime, server_default=text("CURRENT_TIMESTAMP"))


if TYPE_CHECKING:  # pragma: no cover - type checker helper
    from sqlalchemy.engine import Engine
else:  # pragma: no cover - fallback stub

    class Engine:  # type: ignore[too-many-ancestors]
        """Placeholder used when SQLAlchemy is unavailable."""

        pass


SCHEMA = """
CREATE TABLE IF NOT EXISTS forex_rates_rbi (
    rate_date DATE NOT NULL,
    currency TEXT NOT NULL,
    rate REAL NOT NULL,
    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
    PRIMARY KEY(rate_date, currency)
);

CREATE TABLE IF NOT EXISTS forex_rates_sbi (
    rate_date DATE NOT NULL,
    currency TEXT NOT NULL,
    rate REAL NOT NULL,
    tt_buy REAL,
    tt_sell REAL,
    bill_buy REAL,
    bill_sell REAL,
    travel_card_buy REAL,
    travel_card_sell REAL,
    cn_buy REAL,
    cn_sell REAL,
    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP,
    PRIMARY KEY(rate_date, currency)
);

CREATE TABLE IF NOT EXISTS ingestion_metadata (
    source TEXT PRIMARY KEY,
    last_ingested_date DATE NOT NULL,
    updated_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
);

CREATE TABLE IF NOT EXISTS lme_copper_rates (
    rate_date DATE PRIMARY KEY,
    price REAL,
    price_3_month REAL,
    stock INTEGER,
    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
);

CREATE TABLE IF NOT EXISTS lme_aluminum_rates (
    rate_date DATE PRIMARY KEY,
    price REAL,
    price_3_month REAL,
    stock INTEGER,
    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
);
"""

INSERT_RBI_IGNORE_STATEMENT = """
INSERT OR IGNORE INTO forex_rates_rbi(rate_date, currency, rate)
VALUES(?, ?, ?);
"""

INSERT_SBI_IGNORE_STATEMENT = """
INSERT OR IGNORE INTO forex_rates_sbi(
    rate_date,
    currency,
    rate,
    tt_buy,
    tt_sell,
    bill_buy,
    bill_sell,
    travel_card_buy,
    travel_card_sell,
    cn_buy,
    cn_sell
)
VALUES(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?);
"""

UPDATE_RBI_STATEMENT = """
UPDATE forex_rates_rbi
SET rate = ?
WHERE rate_date = ? AND currency = ?;
"""

UPDATE_SBI_STATEMENT = """
UPDATE forex_rates_sbi
SET rate = ?,
    tt_buy = ?,
    tt_sell = ?,
    bill_buy = ?,
    bill_sell = ?,
    travel_card_buy = ?,
    travel_card_sell = ?,
    cn_buy = ?,
    cn_sell = ?
WHERE rate_date = ? AND currency = ?;
"""

INSERT_LME_COPPER_STATEMENT = """
INSERT OR REPLACE INTO lme_copper_rates(
    rate_date,
    price,
    price_3_month,
    stock
)
VALUES(?, ?, ?, ?);
"""

INSERT_LME_ALUMINUM_STATEMENT = """
INSERT OR REPLACE INTO lme_aluminum_rates(
    rate_date,
    price,
    price_3_month,
    stock
)
VALUES(?, ?, ?, ?);
"""


@dataclass(slots=True)
class PersistenceResult:
    """Represents how many rows were inserted or updated in a batch."""

    inserted: int = 0
    updated: int = 0

    @property
    def total(self) -> int:
        """Return the total number of affected rows."""

        return self.inserted + self.updated


class _BackendProtocol(Protocol):
    def insert_rates(self, rows: Sequence[ForexRateRecord]) -> PersistenceResult:
        ...  # pragma: no cover - protocol definition

    def fetch_all(self) -> list[ForexRateRecord]:
        ...  # pragma: no cover - protocol definition

    def fetch_range(
        self, start: date | None = None, end: date | None = None, *, source: str | None = None
    ) -> list[ForexRateRecord]:
        ...  # pragma: no cover - protocol definition

    def close(self) -> None:
        ...  # pragma: no cover - protocol definition

    def latest_rate_date(self, source: str) -> date | None:
        ...  # pragma: no cover - protocol definition

    def ingestion_checkpoint(self, source: str) -> date | None:
        ...  # pragma: no cover - protocol definition

    def update_ingestion_checkpoint(self, source: str, rate_date: date) -> None:
        ...  # pragma: no cover - protocol definition

    def insert_lme_rates(self, metal: str, rows: Sequence[LmeRateRecord]) -> PersistenceResult:
        ...  # pragma: no cover - protocol definition

    def fetch_lme_range(
        self, metal: str, start: date | None = None, end: date | None = None
    ) -> list[LmeRateRecord]:
        ...  # pragma: no cover - protocol definition


class _SQLAlchemyBackend:
    """SQLAlchemy implementation used when the dependency is available."""

    def __init__(self, db_path: str | Path) -> None:
        if not SQLALCHEMY_AVAILABLE:  # pragma: no cover - defensive guard
            raise ModuleNotFoundError("SQLAlchemy is required for SQLiteManager")
        self.db_path = Path(db_path).expanduser().resolve()
        self.engine: Engine = create_engine(
            f"sqlite:///{self.db_path}",
            echo=False,
            future=True,
            connect_args={"check_same_thread": False},
        )
        Base.metadata.create_all(self.engine)
        self._ensure_lme_schema()
        self._SessionFactory: sessionmaker[Session] = sessionmaker(
            bind=self.engine, expire_on_commit=False, future=True
        )

    def _ensure_lme_schema(self) -> None:
        """Ensure older SQLite files include LME columns introduced later."""

        def _missing_columns(table: str, expected: dict[str, str]) -> dict[str, str]:
            with self.engine.connect() as connection:
                result = connection.execute(text(f"PRAGMA table_info({table})"))
                existing = {row[1] for row in result}
            return {
                name: column_type for name, column_type in expected.items() if name not in existing
            }

        def _rebuild_table(table: str) -> None:
            desired = ["rate_date", "price", "price_3_month", "stock", "created_at"]
            columns_csv = ", ".join(desired)
            temp_table = f"{table}_tmp"
            with self.engine.begin() as connection:
                connection.execute(
                    text(
                        f"""
                        CREATE TABLE {temp_table} (
                            rate_date DATE PRIMARY KEY,
                            price REAL,
                            price_3_month REAL,
                            stock INTEGER,
                            created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
                        )
                        """
                    )
                )
                connection.execute(
                    text(
                        f"""
                        INSERT INTO {temp_table} ({columns_csv})
                        SELECT {columns_csv}
                        FROM {table}
                        """
                    )
                )
                connection.execute(text(f"DROP TABLE {table}"))
                connection.execute(text(f"ALTER TABLE {temp_table} RENAME TO {table}"))

        lme_columns = {
            "price": "REAL",
            "price_3_month": "REAL",
            "stock": "INTEGER",
            "created_at": "TIMESTAMP",
        }
        unwanted_columns = {"usd_price", "eur_price", "usd_change", "eur_change"}
        for table in ("lme_copper_rates", "lme_aluminum_rates"):
            missing = _missing_columns(table, lme_columns)
            if missing:
                with self.engine.begin() as connection:
                    for column_name, column_type in missing.items():
                        connection.execute(
                            text(f"ALTER TABLE {table} ADD COLUMN {column_name} {column_type}")
                        )
            with self.engine.connect() as connection:
                result = connection.execute(text(f"PRAGMA table_info({table})"))
                existing = {row[1] for row in result}
            if unwanted_columns.intersection(existing):
                _rebuild_table(table)

    @staticmethod
    def _resolve_lme_model(metal: str):
        normalised = metal.upper()
        if normalised in {"CU", "COPPER"}:
            return _LmeCopperRate
        if normalised in {"AL", "ALUMINUM", "ALUMINIUM"}:
            return _LmeAluminumRate
        raise ValueError(f"Unsupported LME metal: {metal}")

    def insert_rates(self, rows: Sequence[ForexRateRecord]) -> PersistenceResult:
        result = PersistenceResult()
        with self._SessionFactory() as session:
            for row in rows:
                source = (row.source or "RBI").upper()
                pk = {"rate_date": row.rate_date, "currency": row.currency}
                if source == "SBI":
                    sbi_existing = session.get(_SbiRate, pk)
                    if sbi_existing is None:
                        session.add(
                            _SbiRate(
                                rate_date=row.rate_date,
                                currency=row.currency,
                                rate=row.rate,
                                tt_buy=row.tt_buy,
                                tt_sell=row.tt_sell,
                                bill_buy=row.bill_buy,
                                bill_sell=row.bill_sell,
                                travel_card_buy=row.travel_card_buy,
                                travel_card_sell=row.travel_card_sell,
                                cn_buy=row.cn_buy,
                                cn_sell=row.cn_sell,
                            )
                        )
                        result.inserted += 1
                    else:
                        setattr(sbi_existing, "rate", row.rate)
                        setattr(sbi_existing, "tt_buy", row.tt_buy)
                        setattr(sbi_existing, "tt_sell", row.tt_sell)
                        setattr(sbi_existing, "bill_buy", row.bill_buy)
                        setattr(sbi_existing, "bill_sell", row.bill_sell)
                        setattr(sbi_existing, "travel_card_buy", row.travel_card_buy)
                        setattr(sbi_existing, "travel_card_sell", row.travel_card_sell)
                        setattr(sbi_existing, "cn_buy", row.cn_buy)
                        setattr(sbi_existing, "cn_sell", row.cn_sell)
                        result.updated += 1
                else:
                    rbi_existing = session.get(_RbiRate, pk)
                    if rbi_existing is None:
                        session.add(
                            _RbiRate(
                                rate_date=row.rate_date,
                                currency=row.currency,
                                rate=row.rate,
                            )
                        )
                        result.inserted += 1
                    else:
                        setattr(rbi_existing, "rate", row.rate)
                        result.updated += 1
            session.commit()
        return result

    def insert_lme_rates(self, metal: str, rows: Sequence[LmeRateRecord]) -> PersistenceResult:
        model = self._resolve_lme_model(metal)
        result = PersistenceResult()
        if not rows:
            return result
        with self._SessionFactory() as session:
            for row in rows:
                pk = {"rate_date": row.rate_date}
                existing = session.get(model, pk)
                if existing is None:
                    session.add(
                        model(
                            rate_date=row.rate_date,
                            price=row.price,
                            price_3_month=row.price_3_month,
                            stock=row.stock,
                        )
                    )
                    result.inserted += 1
                else:
                    setattr(existing, "price", row.price)
                    setattr(existing, "price_3_month", row.price_3_month)
                    setattr(existing, "stock", row.stock)
                    result.updated += 1
            session.commit()
        return result

    def fetch_all(self) -> list[ForexRateRecord]:
        return self.fetch_range()

    def fetch_range(
        self,
        start: date | None = None,
        end: date | None = None,
        *,
        source: str | None = None,
    ) -> list[ForexRateRecord]:
        records: list[ForexRateRecord] = []
        with self._SessionFactory() as session:
            if source is None or source.upper() == "SBI":
                sbi_stmt = select(_SbiRate).order_by(_SbiRate.rate_date)
                if start is not None:
                    sbi_stmt = sbi_stmt.where(_SbiRate.rate_date >= start)
                if end is not None:
                    sbi_stmt = sbi_stmt.where(_SbiRate.rate_date <= end)
                for sbi_row in session.execute(sbi_stmt).scalars():
                    sbi_model = cast(_SbiRate, sbi_row)
                    records.append(
                        ForexRateRecord(
                            rate_date=cast(date, sbi_model.rate_date),
                            currency=cast(str, sbi_model.currency),
                            rate=cast(float, sbi_model.rate),
                            source="SBI",
                            tt_buy=cast(float | None, sbi_model.tt_buy),
                            tt_sell=cast(float | None, sbi_model.tt_sell),
                            bill_buy=cast(float | None, sbi_model.bill_buy),
                            bill_sell=cast(float | None, sbi_model.bill_sell),
                            travel_card_buy=cast(float | None, sbi_model.travel_card_buy),
                            travel_card_sell=cast(float | None, sbi_model.travel_card_sell),
                            cn_buy=cast(float | None, sbi_model.cn_buy),
                            cn_sell=cast(float | None, sbi_model.cn_sell),
                        )
                    )
        if source is None or source.upper() == "RBI":
            rbi_stmt = select(_RbiRate).order_by(_RbiRate.rate_date)
            if start is not None:
                rbi_stmt = rbi_stmt.where(_RbiRate.rate_date >= start)
            if end is not None:
                rbi_stmt = rbi_stmt.where(_RbiRate.rate_date <= end)
            for rbi_row in session.execute(rbi_stmt).scalars():
                rbi_model = cast(_RbiRate, rbi_row)
                records.append(
                    ForexRateRecord(
                        rate_date=cast(date, rbi_model.rate_date),
                        currency=cast(str, rbi_model.currency),
                        rate=cast(float, rbi_model.rate),
                        source="RBI",
                    )
                )
        return records

    def fetch_lme_range(
        self, metal: str, start: date | None = None, end: date | None = None
    ) -> list[LmeRateRecord]:
        model = self._resolve_lme_model(metal)
        records: list[LmeRateRecord] = []
        with self._SessionFactory() as session:
            stmt = select(model).order_by(model.rate_date)
            if start is not None:
                stmt = stmt.where(model.rate_date >= start)
            if end is not None:
                stmt = stmt.where(model.rate_date <= end)
            for db_row in session.execute(stmt).scalars():
                records.append(
                    LmeRateRecord(
                        rate_date=cast(date, db_row.rate_date),
                        price=cast(float | None, db_row.price),
                        price_3_month=cast(float | None, db_row.price_3_month),
                        stock=cast(int | None, db_row.stock),
                        metal="COPPER" if model is _LmeCopperRate else "ALUMINUM",
                    )
                )
        return records

    def latest_rate_date(self, source: str) -> date | None:
        stmt = select(
            func.max(_SbiRate.rate_date)
            if source.upper() == "SBI"
            else func.max(_RbiRate.rate_date)
        )
        with self._SessionFactory() as session:
            result = session.execute(stmt).scalar_one_or_none()
        return cast(date | None, result)

    def ingestion_checkpoint(self, source: str) -> date | None:
        with self._SessionFactory() as session:
            record = session.get(_IngestionMetadata, source.upper())
            return cast(date | None, record.last_ingested_date) if record else None

    def update_ingestion_checkpoint(self, source: str, rate_date: date) -> None:
        with self._SessionFactory() as session:
            existing = session.get(_IngestionMetadata, source.upper())
            if existing is None:
                session.add(_IngestionMetadata(source=source.upper(), last_ingested_date=rate_date))
            elif existing.last_ingested_date < rate_date:
                cast(Any, existing).last_ingested_date = rate_date
            else:
                return
            session.commit()

    def close(self) -> None:  # pragma: no cover - trivial
        self.engine.dispose()


class _SQLiteFallbackBackend:
    """Legacy sqlite3 implementation kept for environments without SQLAlchemy."""

    def __init__(self, db_path: str | Path) -> None:
        self.db_path = Path(db_path)
        self._connection = sqlite3.connect(self.db_path)
        self._connection.execute("PRAGMA foreign_keys = ON")
        self._connection.row_factory = sqlite3.Row
        self._connection.executescript(SCHEMA)
        self._drop_unwanted_lme_columns()
        self._connection.commit()

    def _drop_unwanted_lme_columns(self) -> None:
        unwanted = {"usd_price", "eur_price", "usd_change", "eur_change"}
        desired = ["rate_date", "price", "price_3_month", "stock", "created_at"]
        columns_csv = ", ".join(desired)
        for table in ("lme_copper_rates", "lme_aluminum_rates"):
            cursor = self._connection.execute(f"PRAGMA table_info({table})")
            existing = {row["name"] for row in cursor.fetchall()}
            if not unwanted.intersection(existing):
                continue
            temp_table = f"{table}_tmp"
            self._connection.execute(
                f"""
                CREATE TABLE {temp_table} (
                    rate_date DATE PRIMARY KEY,
                    price REAL,
                    price_3_month REAL,
                    stock INTEGER,
                    created_at TIMESTAMP DEFAULT CURRENT_TIMESTAMP
                )
                """
            )
            self._connection.execute(
                f"""
                INSERT INTO {temp_table} ({columns_csv})
                SELECT {columns_csv}
                FROM {table}
                """
            )
            self._connection.execute(f"DROP TABLE {table}")
            self._connection.execute(f"ALTER TABLE {temp_table} RENAME TO {table}")

    @staticmethod
    def _resolve_lme_table(metal: str) -> tuple[str, str]:
        normalised = metal.upper()
        if normalised in {"CU", "COPPER"}:
            return "lme_copper_rates", INSERT_LME_COPPER_STATEMENT
        if normalised in {"AL", "ALUMINUM", "ALUMINIUM"}:
            return "lme_aluminum_rates", INSERT_LME_ALUMINUM_STATEMENT
        raise ValueError(f"Unsupported LME metal: {metal}")

    def insert_rates(self, rows: Sequence[ForexRateRecord]) -> PersistenceResult:
        result = PersistenceResult()
        with self._connection:
            for row in rows:
                if (row.source or "RBI").upper() == "SBI":
                    inserted = self._connection.execute(
                        INSERT_SBI_IGNORE_STATEMENT,
                        (
                            row.rate_date.isoformat(),
                            row.currency,
                            row.rate,
                            row.tt_buy,
                            row.tt_sell,
                            row.bill_buy,
                            row.bill_sell,
                            row.travel_card_buy,
                            row.travel_card_sell,
                            row.cn_buy,
                            row.cn_sell,
                        ),
                    ).rowcount
                    if inserted:
                        result.inserted += inserted
                        continue
                    updated = self._connection.execute(
                        UPDATE_SBI_STATEMENT,
                        (
                            row.rate,
                            row.tt_buy,
                            row.tt_sell,
                            row.bill_buy,
                            row.bill_sell,
                            row.travel_card_buy,
                            row.travel_card_sell,
                            row.cn_buy,
                            row.cn_sell,
                            row.rate_date.isoformat(),
                            row.currency,
                        ),
                    ).rowcount
                    result.updated += updated
                else:
                    inserted = self._connection.execute(
                        INSERT_RBI_IGNORE_STATEMENT,
                        (
                            row.rate_date.isoformat(),
                            row.currency,
                            row.rate,
                        ),
                    ).rowcount
                    if inserted:
                        result.inserted += inserted
                        continue
                    updated = self._connection.execute(
                        UPDATE_RBI_STATEMENT,
                        (
                            row.rate,
                            row.rate_date.isoformat(),
                            row.currency,
                        ),
                    ).rowcount
                    result.updated += updated
        return result

    def insert_lme_rates(self, metal: str, rows: Sequence[LmeRateRecord]) -> PersistenceResult:
        table, statement = self._resolve_lme_table(metal)
        result = PersistenceResult()
        if not rows:
            return result
        with self._connection:
            for row in rows:
                inserted = self._connection.execute(
                    statement,
                    (
                        row.rate_date.isoformat(),
                        row.price,
                        row.price_3_month,
                        row.stock,
                    ),
                ).rowcount
                if inserted:
                    result.inserted += inserted
                else:
                    updated = self._connection.execute(
                        f"""
                        UPDATE {table}
                        SET price = ?,
                            price_3_month = ?,
                            stock = ?
                        WHERE rate_date = ?
                        """,
                        (
                            row.price,
                            row.price_3_month,
                            row.stock,
                            row.rate_date.isoformat(),
                        ),
                    ).rowcount
                    result.updated += updated
        return result

    def fetch_all(self) -> list[ForexRateRecord]:
        return self.fetch_range()

    def fetch_range(
        self,
        start: date | None = None,
        end: date | None = None,
        *,
        source: str | None = None,
    ) -> list[ForexRateRecord]:
        def _build_query(table: str) -> tuple[str, list[str]]:
            clauses: list[str] = []
            params: list[str] = []
            if start is not None:
                clauses.append("rate_date >= ?")
                params.append(start.isoformat())
            if end is not None:
                clauses.append("rate_date <= ?")
                params.append(end.isoformat())
            where = ""
            if clauses:
                where = " WHERE " + " AND ".join(clauses)
            return (
                f"SELECT * FROM {table}{where} ORDER BY rate_date",
                params,
            )

        records: list[ForexRateRecord] = []

        if source is None or source.upper() == "SBI":
            query, params = _build_query("forex_rates_sbi")
            for row in self._connection.execute(query, params).fetchall():
                records.append(
                    ForexRateRecord(
                        rate_date=date.fromisoformat(row["rate_date"]),
                        currency=row["currency"],
                        rate=row["rate"],
                        source="SBI",
                        tt_buy=row["tt_buy"],
                        tt_sell=row["tt_sell"],
                        bill_buy=row["bill_buy"],
                        bill_sell=row["bill_sell"],
                        travel_card_buy=row["travel_card_buy"],
                        travel_card_sell=row["travel_card_sell"],
                        cn_buy=row["cn_buy"],
                        cn_sell=row["cn_sell"],
                    )
                )

        if source is None or source.upper() == "RBI":
            query, params = _build_query("forex_rates_rbi")
            for row in self._connection.execute(query, params).fetchall():
                records.append(
                    ForexRateRecord(
                        rate_date=date.fromisoformat(row["rate_date"]),
                        currency=row["currency"],
                        rate=row["rate"],
                        source="RBI",
                    )
                )

        return records

    def fetch_lme_range(
        self, metal: str, start: date | None = None, end: date | None = None
    ) -> list[LmeRateRecord]:
        table, _ = self._resolve_lme_table(metal)

        def _build_query() -> tuple[str, list[str]]:
            clauses: list[str] = []
            params: list[str] = []
            if start is not None:
                clauses.append("rate_date >= ?")
                params.append(start.isoformat())
            if end is not None:
                clauses.append("rate_date <= ?")
                params.append(end.isoformat())
            where = ""
            if clauses:
                where = " WHERE " + " AND ".join(clauses)
            return (f"SELECT * FROM {table}{where} ORDER BY rate_date", params)

        query, params = _build_query()
        records: list[LmeRateRecord] = []
        for row in self._connection.execute(query, params).fetchall():
            records.append(
                LmeRateRecord(
                    rate_date=date.fromisoformat(row["rate_date"]),
                    price=row["price"],
                    price_3_month=row["price_3_month"],
                    stock=row["stock"],
                    metal="COPPER" if table.endswith("copper_rates") else "ALUMINUM",
                )
            )
        return records

    def latest_rate_date(self, source: str) -> date | None:
        table = "forex_rates_sbi" if source.upper() == "SBI" else "forex_rates_rbi"
        cursor = self._connection.execute(f"SELECT MAX(rate_date) AS latest FROM {table}")
        value = cursor.fetchone()["latest"]
        return date.fromisoformat(value) if value else None

    def ingestion_checkpoint(self, source: str) -> date | None:
        cursor = self._connection.execute(
            "SELECT last_ingested_date FROM ingestion_metadata WHERE source = ?",
            (source.upper(),),
        )
        value = cursor.fetchone()
        if value is None:
            return None
        return date.fromisoformat(value["last_ingested_date"])

    def update_ingestion_checkpoint(self, source: str, rate_date: date) -> None:
        with self._connection:
            self._connection.execute(
                """
                INSERT INTO ingestion_metadata(source, last_ingested_date)
                VALUES(?, ?)
                ON CONFLICT(source) DO UPDATE SET last_ingested_date = excluded.last_ingested_date
                WHERE excluded.last_ingested_date > ingestion_metadata.last_ingested_date
                """,
                (source.upper(), rate_date.isoformat()),
            )

    def close(self) -> None:  # pragma: no cover - trivial
        self._connection.close()


class SQLiteManager:
    """Facade that prefers SQLAlchemy but gracefully falls back to sqlite3."""

    def __init__(self, db_path: str | Path = DEFAULT_SQLITE_DB_PATH) -> None:
        self.db_path = Path(db_path)
        self._backend: _BackendProtocol
        if SQLALCHEMY_AVAILABLE:
            LOGGER.info("Using SQLAlchemy backend for SQLiteManager")
            self._backend = _SQLAlchemyBackend(self.db_path)
        else:
            LOGGER.warning("SQLAlchemy not installed; using sqlite3 fallback backend")
            self._backend = _SQLiteFallbackBackend(self.db_path)

    def insert_rates(self, rows: Sequence[ForexRateRecord]) -> PersistenceResult:
        result = self._backend.insert_rates(rows)
        LOGGER.info(
            "Inserted %s rows, updated %s rows (total %s)",
            result.inserted,
            result.updated,
            result.total,
        )
        return result

    def fetch_all(self) -> list[ForexRateRecord]:
        return self._backend.fetch_all()

    def fetch_range(
        self,
        start: date | None = None,
        end: date | None = None,
        *,
        source: str | None = None,
    ) -> list[ForexRateRecord]:
        return self._backend.fetch_range(start, end, source=source)

    def insert_lme_rates(self, metal: str, rows: Sequence[LmeRateRecord]) -> PersistenceResult:
        result = self._backend.insert_lme_rates(metal, rows)
        LOGGER.info(
            "Inserted %s %s rows, updated %s rows (total %s)",
            result.inserted,
            metal,
            result.updated,
            result.total,
        )
        if rows:
            latest_day = max(row.rate_date for row in rows)
            self._backend.update_ingestion_checkpoint(
                f"LME_{_normalise_lme_metal(metal)}", latest_day
            )
        return result

    def fetch_lme_range(
        self, metal: str, start: date | None = None, end: date | None = None
    ) -> list[LmeRateRecord]:
        return self._backend.fetch_lme_range(metal, start, end)

    def close(self) -> None:  # pragma: no cover - trivial
        self._backend.close()

    def latest_rate_date(self, source: str) -> date | None:
        return self._backend.latest_rate_date(source)

    def ingestion_checkpoint(self, source: str) -> date | None:
        return self._backend.ingestion_checkpoint(source)

    def update_ingestion_checkpoint(self, source: str, rate_date: date) -> None:
        self._backend.update_ingestion_checkpoint(source, rate_date)

    def __enter__(self) -> "SQLiteManager":  # pragma: no cover - trivial
        return self

    def __exit__(self, exc_type, exc, tb) -> None:  # pragma: no cover - trivial
        self.close()


def _normalise_lme_metal(metal: str) -> str:
    upper = metal.upper()
    if upper in {"CU", "COPPER"}:
        return "COPPER"
    if upper in {"AL", "ALUMINUM", "ALUMINIUM"}:
        return "ALUMINUM"
    return upper



"""Tests for the public package facade."""

from datetime import date, timedelta
from pathlib import Path
from typing import Any

import pytest

from fx_bharat import DatabaseBackend, DatabaseConnectionInfo, FxBharat, __version__
from fx_bharat.db.mongo_backend import MongoBackend
from fx_bharat.db.sqlite_manager import PersistenceResult
from fx_bharat.ingestion.models import ForexRateRecord, LmeRateRecord
from fx_bharat.utils.rbi import RBI_MIN_AVAILABLE_DATE


def test_fx_bharat_class_is_exposed() -> None:
    """The package should expose the FxBharat class."""

    assert FxBharat.__version__ == __version__


def test_fx_bharat_defaults_to_sqlite() -> None:
    fx_bharat = FxBharat()

    assert fx_bharat.connection_info.backend is DatabaseBackend.SQLITE
    assert fx_bharat.sqlite_manager is not None


def test_fx_bharat_supports_external_backends() -> None:
    fx_bharat = FxBharat(db_config="mysql://user:pwd@localhost:3306/forex")

    assert fx_bharat.connection_info.backend is DatabaseBackend.MYSQL
    assert fx_bharat.connection_info.username == "user"
    assert fx_bharat.connection_info.password == "pwd"
    assert fx_bharat.connection_info.host == "localhost"
    assert fx_bharat.connection_info.port == 3306
    assert fx_bharat.sqlite_manager is None


@pytest.mark.parametrize(
    "url, backend",
    [
        ("postgresql://localhost/forex", DatabaseBackend.POSTGRES),
        ("postgres://localhost/forex", DatabaseBackend.POSTGRES),
        ("mongodb://localhost:27017/forex", DatabaseBackend.MONGODB),
    ],
)
def test_fx_bharat_normalises_backend_names(url: str, backend: DatabaseBackend) -> None:
    fx_bharat = FxBharat(db_config=url)

    assert fx_bharat.connection_info.backend is backend


@pytest.mark.parametrize(
    "scheme, backend",
    [
        ("postgresql", DatabaseBackend.POSTGRES),
        ("postgres", DatabaseBackend.POSTGRES),
        ("postgressql", DatabaseBackend.POSTGRES),
        ("mysql+pymysql", DatabaseBackend.MYSQL),
        ("sqlite", DatabaseBackend.SQLITE),
        ("mongodb+srv", DatabaseBackend.MONGODB),
    ],
)
def test_database_backend_from_scheme_handles_aliases(
    scheme: str, backend: DatabaseBackend
) -> None:
    assert DatabaseBackend.from_scheme(scheme) is backend


def test_fx_bharat_validates_urls() -> None:
    with pytest.raises(ValueError):
        FxBharat(db_config="localhost:3306/forex")


def test_database_connection_info_requires_scheme() -> None:
    with pytest.raises(ValueError, match="DB_URL must include a scheme"):
        DatabaseConnectionInfo.from_url("localhost/forex")


def test_database_connection_info_from_url() -> None:
    info = DatabaseConnectionInfo.from_url("postgres://user:pwd@db:5432/app")

    assert info.backend is DatabaseBackend.POSTGRES
    assert info.username == "user"
    assert info.password == "pwd"
    assert info.host == "db"
    assert info.port == 5432
    assert info.name == "app"


def test_database_name_is_extracted_from_query_parameters() -> None:
    info = DatabaseConnectionInfo.from_url(
        "mongodb://user:pwd@cluster0.example.com/?retryWrites=false&w=majority&appName=Cluster0DATABASE_NAME=test",
    )

    assert info.backend is DatabaseBackend.MONGODB
    assert info.name == "test"
    assert info.url.startswith(
        "mongodb://user:pwd@cluster0.example.com/test?retryWrites=false&w=majority&appName=Cluster0"
    )


def test_mongodb_srv_connections_are_supported() -> None:
    info = DatabaseConnectionInfo.from_url(
        "mongodb+srv://127.0.0.1:27017/?DATABASE_NAME=forex&retryWrites=false&w=majority&appName=Cluster0",
    )

    assert info.backend is DatabaseBackend.MONGODB
    assert info.name == "forex"
    assert info.url.startswith(
        "mongodb+srv://127.0.0.1:27017/forex?retryWrites=false&w=majority&appName=Cluster0"
    )


@pytest.mark.parametrize(
    "dsn",
    [
        "postgresql+asyncpg://postgres:postgres@localhost/forex",
        "postgressql+asyncpg://postgres:postgres@localhost/forex",
    ],
)
def test_postgres_asyncpg_urls_are_normalised(dsn: str) -> None:
    info = DatabaseConnectionInfo.from_url(dsn)

    assert info.backend is DatabaseBackend.POSTGRES
    assert info.url.startswith("postgresql://postgres:postgres@localhost/forex")


def test_database_connection_info_uses_database_name_query_parameter() -> None:
    info = DatabaseConnectionInfo.from_url("mysql://localhost/?DATABASE_NAME=forex")

    assert info.backend is DatabaseBackend.MYSQL
    assert info.name == "forex"
    assert info.url.startswith("mysql://localhost/forex")


def test_fx_bharat_accepts_prebuilt_config() -> None:
    config = DatabaseConnectionInfo.from_url("mongodb://example.com:27017/fx")

    fx_bharat = FxBharat(db_config=config)

    assert fx_bharat.connection_info is config
    assert fx_bharat.connection_info.backend is DatabaseBackend.MONGODB


@pytest.fixture()
def sqlite_fx(tmp_path: Path) -> FxBharat:
    db_path = tmp_path / "forex.db"
    config = DatabaseConnectionInfo(
        backend=DatabaseBackend.SQLITE,
        url=f"sqlite:///{db_path}",
        name=str(db_path),
        username=None,
        password=None,
        host=None,
        port=None,
    )
    return FxBharat(db_config=config)


def test_connection_reports_success_for_sqlite() -> None:
    fx_bharat = FxBharat()

    success, error = fx_bharat.connection()

    assert success is True
    assert error is None


def test_connection_reports_failure_for_unreachable_external_db() -> None:
    external = FxBharat(db_config="postgres://user:pass@db.example.com:5432/forex")

    success, error = external.connection()

    assert success is False
    assert isinstance(error, str)
    assert error


def test_connection_reports_missing_driver(monkeypatch: pytest.MonkeyPatch) -> None:
    external = FxBharat(db_config="postgres://user:pass@db.example.com:5432/forex")

    def _missing_driver(*_: Any, **__: Any) -> None:
        raise ModuleNotFoundError("psycopg2")

    monkeypatch.setattr("fx_bharat.create_engine", _missing_driver)

    success, error = external.connection()

    assert success is False
    assert "psycopg2" in (error or "")
    assert "pip install psycopg2-binary" in (error or "")


def test_connection_uses_pymongo_for_mongodb(monkeypatch: pytest.MonkeyPatch) -> None:
    external = FxBharat(db_config="mongodb://user:pass@cluster0.example.com:27017/forex")

    class DummyClient:
        def __init__(self, url: str, serverSelectionTimeoutMS: int) -> None:  # noqa: N803
            assert url.startswith("mongodb://")
            assert serverSelectionTimeoutMS == 5000
            self.closed = False
            self.admin = self

        def command(self, name: str) -> None:
            assert name == "ping"

        def close(self) -> None:
            self.closed = True

    monkeypatch.setattr("fx_bharat.MongoClient", DummyClient)

    success, error = external.connection()

    assert success is True
    assert error is None


def test_history_lme_returns_snapshots(sqlite_fx: FxBharat) -> None:
    assert sqlite_fx.sqlite_manager is not None
    sqlite_fx.sqlite_manager.insert_lme_rates(
        "COPPER",
        [
            LmeRateRecord(
                rate_date=date(2024, 1, 1),
                price=8500.0,
                price_3_month=8450.0,
                stock=100,
                metal="COPPER",
            )
        ],
    )
    sqlite_fx.sqlite_manager.insert_lme_rates(
        "ALUMINUM",
        [
            LmeRateRecord(
                rate_date=date(2024, 1, 2),
                price=2500.0,
                price_3_month=2450.0,
                stock=200,
                metal="ALUMINUM",
            )
        ],
    )

    snapshots = sqlite_fx.history_lme(date(2024, 1, 1), date(2024, 1, 3))

    assert {snap["metal"] for snap in snapshots} == {"COPPER", "ALUMINUM"}
    assert {snap["rate_date"] for snap in snapshots} == {date(2024, 1, 1), date(2024, 1, 2)}


def test_history_lme_rejects_invalid_filter(sqlite_fx: FxBharat) -> None:
    with pytest.raises(ValueError):
        sqlite_fx.history_lme(date(2024, 1, 1), date(2024, 1, 2), source_filter="gold")


def test_mongodb_connection_reports_missing_driver(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    external = FxBharat(db_config="mongodb://user:pass@cluster0.example.com:27017/forex")

    monkeypatch.setattr("fx_bharat.MongoClient", None)

    success, error = external.connection()

    assert success is False
    assert error is not None
    assert "pymongo" in error


@pytest.mark.parametrize(
    "url, attr_name, expected_kwargs",
    [
        ("postgresql://user:pass@db.example.com:5432/forex", "PostgresBackend", {}),
        ("mysql://user:pass@db.example.com:3306/forex", "MySQLBackend", {}),
        (
            "mongodb://user:pass@db.example.com:27017/forex",
            "MongoBackend",
            {"database": "forex"},
        ),
    ],
)
def test_build_external_backend_instantiates_strategy(
    monkeypatch: pytest.MonkeyPatch,
    url: str,
    attr_name: str,
    expected_kwargs: dict[str, str],
) -> None:
    captured: dict[str, object] = {}

    class DummyStrategy:
        def __init__(self, url: str, **kwargs: object) -> None:
            captured["url"] = url
            captured["kwargs"] = kwargs

    monkeypatch.setattr(f"fx_bharat.{attr_name}", DummyStrategy)
    fx_bharat = FxBharat(db_config=url)

    strategy = fx_bharat._build_external_backend()

    assert isinstance(strategy, DummyStrategy)
    assert captured["url"] == fx_bharat.connection_info.url
    assert captured["kwargs"] == expected_kwargs


def test_get_backend_strategy_raises_for_uninitialised_sqlite() -> None:
    fx_bharat = FxBharat()
    fx_bharat._backend_strategy = None

    with pytest.raises(RuntimeError):
        fx_bharat._get_backend_strategy()


def test_get_backend_strategy_caches_external_backend(monkeypatch: pytest.MonkeyPatch) -> None:
    fx_bharat = FxBharat(db_config="postgresql://user:pass@db:5432/forex")
    sentinel = object()

    calls = {"count": 0}

    def _fake_build(self: FxBharat) -> object:  # noqa: ANN001
        calls["count"] += 1
        return sentinel

    monkeypatch.setattr(FxBharat, "_build_external_backend", _fake_build)

    first = fx_bharat._get_backend_strategy()
    second = fx_bharat._get_backend_strategy()

    assert first is second is sentinel
    assert calls["count"] == 1


def test_uses_inhouse_sqlite_reports_backend() -> None:
    sqlite_app = FxBharat()
    mysql_app = FxBharat(db_config="mysql://user:pass@db.example.com:3306/forex")

    assert sqlite_app.uses_inhouse_sqlite() is True
    assert mysql_app.uses_inhouse_sqlite() is False


def test_mongodb_backend_uses_bulk_writes(monkeypatch: pytest.MonkeyPatch) -> None:
    backend = MongoBackend.__new__(MongoBackend)

    class DummyUpdateOne:
        def __init__(
            self, filter: dict[str, str], update: dict[str, dict[str, Any]], *, upsert: bool
        ) -> None:
            self.filter = filter
            self.update = update
            self.upsert = upsert

    monkeypatch.setattr("fx_bharat.db.mongo_backend.UpdateOne", DummyUpdateOne)

    class DummyCollection:
        def __init__(self) -> None:
            self.calls: list[tuple[list[object], bool]] = []

        def bulk_write(self, operations: list[object], ordered: bool) -> object:  # noqa: ANN401
            self.calls.append((operations, ordered))

            class Result:
                upserted_count = 0
                modified_count = 0

            return Result()

    dummy_collection = DummyCollection()
    backend._collection = dummy_collection  # type: ignore[attr-defined]
    rows = [
        ForexRateRecord(rate_date=date(2020, 1, 1), currency="USD", rate=1.0, source="RBI"),
        ForexRateRecord(rate_date=date(2020, 1, 2), currency="USD", rate=1.1, source="RBI"),
    ]

    result = backend.insert_rates(rows)

    assert len(dummy_collection.calls) == 1
    operations, ordered = dummy_collection.calls[0]
    assert ordered is False
    assert len(operations) == len(rows)
    assert result.inserted == len(rows)


def test_conection_alias_routes_to_connection() -> None:
    fx_bharat = FxBharat()

    assert fx_bharat.conection() == fx_bharat.connection()


def _seed_sample_rates(fx: FxBharat) -> None:
    assert fx.sqlite_manager is not None
    fx.sqlite_manager.insert_rates(
        [
            ForexRateRecord(rate_date=date(2023, 1, 1), currency="USD", rate=82.0),
            ForexRateRecord(rate_date=date(2023, 1, 1), currency="EUR", rate=88.0),
            ForexRateRecord(rate_date=date(2023, 1, 2), currency="USD", rate=83.0),
            ForexRateRecord(rate_date=date(2023, 1, 8), currency="USD", rate=84.0),
            ForexRateRecord(rate_date=date(2023, 1, 8), currency="EUR", rate=90.0),
            ForexRateRecord(rate_date=date(2023, 2, 5), currency="USD", rate=85.0),
            ForexRateRecord(rate_date=date(2023, 2, 5), currency="EUR", rate=92.0),
        ]
    )


def test_rate_returns_latest_snapshot(sqlite_fx: FxBharat) -> None:
    _seed_sample_rates(sqlite_fx)

    snapshots = sqlite_fx.rate()

    assert len(snapshots) == 1
    snapshot = snapshots[0]
    assert snapshot["rate_date"] == date(2023, 2, 5)
    assert snapshot["base_currency"] == "INR"
    assert snapshot["source"] == "RBI"
    assert snapshot["rates"] == {"EUR": 92.0, "USD": 85.0}


def test_rate_supports_specific_date(sqlite_fx: FxBharat) -> None:
    _seed_sample_rates(sqlite_fx)

    snapshots = sqlite_fx.rate(date(2023, 1, 2))

    assert len(snapshots) == 1
    snapshot = snapshots[0]
    assert snapshot["rate_date"] == date(2023, 1, 2)
    assert snapshot["source"] == "RBI"
    assert snapshot["rates"] == {"USD": 83.0}


def test_rate_returns_both_sources(sqlite_fx: FxBharat) -> None:
    _seed_sample_rates(sqlite_fx)
    assert sqlite_fx.sqlite_manager is not None
    sqlite_fx.sqlite_manager.insert_rates(
        [
            ForexRateRecord(rate_date=date(2023, 3, 1), currency="USD", rate=90.0, source="SBI"),
            ForexRateRecord(rate_date=date(2023, 3, 1), currency="EUR", rate=95.0, source="SBI"),
        ]
    )

    snapshots = sqlite_fx.rate()

    assert [snap["source"] for snap in snapshots] == ["SBI", "RBI"]
    sbi_snapshot, rbi_snapshot = snapshots
    assert sbi_snapshot["rate_date"] == date(2023, 3, 1)
    assert sbi_snapshot["rates"]["USD"]["rate"] == 90.0
    assert sbi_snapshot["rates"]["EUR"]["tt_sell"] is None or isinstance(
        sbi_snapshot["rates"]["EUR"], dict
    )
    assert rbi_snapshot["source"] == "RBI"


def test_rate_rejects_dates_before_rbi_minimum(sqlite_fx: FxBharat) -> None:
    with pytest.raises(ValueError, match="RBI do not provide the data before 12/04/2022"):
        sqlite_fx.rate(date(2022, 4, 11))


def test_history_supports_frequency(sqlite_fx: FxBharat) -> None:
    _seed_sample_rates(sqlite_fx)

    monthly = sqlite_fx.history(date(2023, 1, 1), date(2023, 2, 28), frequency="monthly")

    assert [entry["rate_date"] for entry in monthly] == [
        date(2023, 1, 8),
        date(2023, 2, 5),
    ]
    assert monthly[0]["rates"]["USD"] == 84.0
    assert monthly[1]["rates"]["EUR"] == 92.0


def test_history_validate_inputs(sqlite_fx: FxBharat) -> None:
    with pytest.raises(ValueError):
        sqlite_fx.history(date(2023, 2, 1), date(2023, 1, 1))
    with pytest.raises(ValueError):
        sqlite_fx.history(date(2023, 1, 1), date(2023, 1, 2), frequency="hourly")  # type: ignore[arg-type]


def test_history_rejects_requests_before_rbi_minimum(sqlite_fx: FxBharat) -> None:
    with pytest.raises(ValueError, match="RBI do not provide the data before 12/04/2022"):
        sqlite_fx.history(date(2022, 4, 11), date(2022, 4, 20))


def test_migrate_requires_external_backend(sqlite_fx: FxBharat) -> None:
    with pytest.raises(ValueError):
        sqlite_fx.migrate()


def test_database_connection_info_uses_database_name_query() -> None:
    info = DatabaseConnectionInfo.from_url(
        "postgres://user:pwd@db.example.com:5432/?DATABASE_NAME=forex&sslmode=prefer",
    )

    assert info.name == "forex"
    assert "DATABASE_NAME" not in info.url
    assert info.url.endswith("/forex?sslmode=prefer")


def test_seed_rejects_inverted_date_range(sqlite_fx: FxBharat) -> None:
    with pytest.raises(ValueError, match="from_date must be on or before to_date"):
        sqlite_fx.seed(from_date=date(2023, 2, 1), to_date=date(2023, 1, 1))


def test_seed_with_explicit_range_targets_rbi(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    db_path = tmp_path / "seed.db"
    fx = FxBharat(db_config=f"sqlite:///{db_path}")

    calls: dict[str, object] = {}

    def _fake_rbi(start: str, end: str, *, incremental: bool, **_: object) -> None:
        calls["rbi"] = (start, end, incremental)

    def _fake_sbi_hist(**_: object) -> None:
        calls["sbi_hist"] = True

    def _fake_sbi_today(**_: object) -> None:
        calls["sbi_today"] = True

    monkeypatch.setattr("fx_bharat.seeds.populate_rbi_forex.seed_rbi_forex", _fake_rbi)
    monkeypatch.setattr("fx_bharat.seeds.populate_sbi_forex.seed_sbi_historical", _fake_sbi_hist)
    monkeypatch.setattr("fx_bharat.seeds.populate_sbi_forex.seed_sbi_today", _fake_sbi_today)

    fx.seed(from_date=date(2023, 1, 1), to_date=date(2023, 1, 3), source="RBI")

    assert calls["rbi"] == ("2023-01-01", "2023-01-03", False)
    assert "sbi_hist" not in calls
    assert "sbi_today" not in calls


def test_seed_default_fetches_sbi_today_and_history(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    db_path = tmp_path / "seed.db"
    fx = FxBharat(db_config=f"sqlite:///{db_path}")

    calls: dict[str, object] = {}
    today = date.today()

    def _fake_rbi(start: str, end: str, **_: object) -> None:
        calls["rbi"] = (start, end)

    def _fake_sbi_hist(*, start, end, **_: object) -> None:  # type: ignore[no-untyped-def]
        calls["sbi_hist"] = (start, end)

    def _fake_sbi_today(**_: object) -> None:
        calls["sbi_today"] = True

    monkeypatch.setattr("fx_bharat.seeds.populate_rbi_forex.seed_rbi_forex", _fake_rbi)
    monkeypatch.setattr("fx_bharat.seeds.populate_sbi_forex.seed_sbi_historical", _fake_sbi_hist)
    monkeypatch.setattr("fx_bharat.seeds.populate_sbi_forex.seed_sbi_today", _fake_sbi_today)

    fx.seed(resource_dir=tmp_path / "resources")

    assert calls["rbi"][1] == today.isoformat()
    assert calls["sbi_hist"][1] == today - timedelta(days=1)
    assert calls["sbi_today"] is True


def test_seed_mirrors_rows_to_external_backend(monkeypatch: pytest.MonkeyPatch) -> None:
    external = FxBharat(db_config="postgres://user:pwd@db:5432/fx")

    calls: dict[str, object] = {}

    def _fake_seed_rbi(
        from_str: str, to_str: str, *, db_path: Path, incremental: bool, **_: object
    ) -> None:
        calls["range"] = (from_str, to_str, incremental)
        calls["db_path"] = Path(db_path)

    def _fake_seed_sbi_today(*, db_path: Path, resource_dir: Path, **_: object) -> None:
        calls["sbi"] = (Path(db_path), resource_dir)

    monkeypatch.setattr("fx_bharat.seeds.populate_rbi_forex.seed_rbi_forex", _fake_seed_rbi)
    monkeypatch.setattr("fx_bharat.seeds.populate_sbi_forex.seed_sbi_today", _fake_seed_sbi_today)

    today = date.today()
    monkeypatch.setattr(
        FxBharat,
        "_get_ingestion_checkpoint",
        lambda self, db_path, source: None,
    )

    created_sqlite_backends: list[object] = []

    class DummySQLiteBackend:
        def __init__(self, db_path: Path | str, manager=None):  # type: ignore[no-untyped-def]
            self.db_path = Path(db_path)
            self.manager = manager
            self.fetch_called_with: list[tuple[date | None, date | None, str | None]] = []
            self.closed = False
            self.rows = [
                ForexRateRecord(rate_date=date.today(), currency="USD", rate=82.1, source="RBI"),
                ForexRateRecord(rate_date=date.today(), currency="EUR", rate=90.2, source="SBI"),
            ]
            created_sqlite_backends.append(self)

        def fetch_range(self, start=None, end=None, source=None):  # type: ignore[no-untyped-def]
            self.fetch_called_with.append((start, end, source))
            return list(self.rows)

        def close(self) -> None:
            self.closed = True

    class DummyExternalBackend:
        def __init__(self) -> None:
            self.rows: list[ForexRateRecord] = []
            self.ensure_called = 0

        def ensure_schema(self) -> None:
            self.ensure_called += 1

        def insert_rates(self, rows):  # type: ignore[no-untyped-def]
            self.rows.extend(rows)

    monkeypatch.setattr("fx_bharat.SQLiteBackend", DummySQLiteBackend)
    external._backend_strategy = DummyExternalBackend()

    external.seed(from_date=today, to_date=today, resource_dir=Path("resources"))

    today_iso = today.isoformat()
    assert calls["range"] == (today_iso, today_iso, False)
    assert isinstance(calls["db_path"], Path)
    sqlite_backend = created_sqlite_backends[0]
    assert (today, today, "RBI") in sqlite_backend.fetch_called_with
    assert (today, today, "SBI") in sqlite_backend.fetch_called_with
    assert sqlite_backend.closed is True

    backend = external._backend_strategy
    assert isinstance(backend, DummyExternalBackend)
    assert backend.ensure_called == 1
    assert len(backend.rows) == 4
    assert {row.currency for row in backend.rows} == {"USD", "EUR"}


def test_seed_inserts_today_for_both_sources(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    db_path = tmp_path / "seed.db"
    config = DatabaseConnectionInfo(
        backend=DatabaseBackend.SQLITE,
        url=f"sqlite:///{db_path}",
        name=str(db_path),
        username=None,
        password=None,
        host=None,
        port=None,
    )
    fx = FxBharat(db_config=config)

    today = date.today().isoformat()
    min_rbi_date = RBI_MIN_AVAILABLE_DATE.isoformat()
    calls: dict[str, object] = {}

    def _fake_seed_rbi(from_str: str, to_str: str, *, db_path: Path, **_: object) -> None:
        calls["rbi"] = (from_str, to_str, Path(db_path))

    def _fake_seed_sbi_today(*, db_path: Path, resource_dir: Path, **_: object) -> None:
        calls["sbi"] = (Path(db_path), resource_dir)

    monkeypatch.setattr("fx_bharat.seeds.populate_rbi_forex.seed_rbi_forex", _fake_seed_rbi)
    monkeypatch.setattr("fx_bharat.seeds.populate_sbi_forex.seed_sbi_today", _fake_seed_sbi_today)

    fx.seed(resource_dir=tmp_path / "resources")

    assert calls["rbi"][:2] == (min_rbi_date, today)
    assert calls["rbi"][2] == db_path
    assert calls["sbi"][0] == db_path
    assert calls["sbi"][1] == tmp_path / "resources"


def test_migrate_copies_rows_to_external_backend(monkeypatch: pytest.MonkeyPatch) -> None:
    class DummySQLiteBackend:
        def __init__(self, path: Path) -> None:
            self.path = path
            self.closed = False

        def fetch_range(
            self,
            start: date | None = None,
            end: date | None = None,
            *,
            source: str | None = None,
        ) -> list[ForexRateRecord]:
            records = [
                ForexRateRecord(
                    rate_date=date(2024, 1, 1),
                    currency="USD",
                    rate=82.5,
                    source="RBI",
                ),
                ForexRateRecord(
                    rate_date=date(2024, 1, 2),
                    currency="EUR",
                    rate=89.1,
                    source="SBI",
                ),
            ]
            if source is None:
                return records
            return [row for row in records if (row.source or "RBI").upper() == source.upper()]

        def fetch_lme_range(  # type: ignore[no-untyped-def]
            self, metal: str, start: date | None = None, end: date | None = None
        ):
            return []

        def close(self) -> None:
            self.closed = True

    class DummyExternalBackend:
        def __init__(self) -> None:
            self.rows: list[ForexRateRecord] = []
            self.ensure_called = 0
            self.checkpoints: dict[str, date] = {}

        def ensure_schema(self) -> None:
            self.ensure_called += 1

        def insert_rates(self, rows: list[ForexRateRecord]) -> PersistenceResult:
            self.rows.extend(rows)
            return PersistenceResult(inserted=len(rows))

        def update_ingestion_checkpoint(self, source: str, rate_date: date) -> None:
            self.checkpoints[source] = rate_date

    monkeypatch.setattr("fx_bharat.SQLiteBackend", DummySQLiteBackend)

    external = FxBharat(db_config="mysql://user:pwd@localhost:3306/fx")
    external._backend_strategy = DummyExternalBackend()

    external.migrate()

    backend = external._backend_strategy
    assert isinstance(backend, DummyExternalBackend)
    assert backend.ensure_called == 1
    assert len(backend.rows) == 2
    assert backend.checkpoints["RBI"] == date(2024, 1, 1)
    assert backend.checkpoints["SBI"] == date(2024, 1, 2)


def test_migrate_respects_date_range(monkeypatch: pytest.MonkeyPatch) -> None:
    created_sqlite_backends: list[object] = []

    class DummySQLiteBackend:
        def __init__(self, path: Path) -> None:
            self.path = path
            self.fetch_calls: list[tuple[date | None, date | None, str | None]] = []
            self.lme_calls: list[tuple[str, date | None, date | None]] = []
            self.closed = False
            created_sqlite_backends.append(self)

        def fetch_range(
            self,
            start: date | None = None,
            end: date | None = None,
            *,
            source: str | None = None,
        ) -> list[ForexRateRecord]:
            self.fetch_calls.append((start, end, source))
            if source is None:
                return []
            return [
                ForexRateRecord(
                    rate_date=start or date(2024, 1, 1),
                    currency="USD",
                    rate=80.0,
                    source=source,
                )
            ]

        def fetch_lme_range(
            self, metal: str, start: date | None = None, end: date | None = None
        ) -> list[LmeRateRecord]:
            self.lme_calls.append((metal, start, end))
            return [
                LmeRateRecord(
                    rate_date=start or date(2024, 1, 1),
                    price=1.0,
                    price_3_month=2.0,
                    stock=3,
                    metal=metal,
                )
            ]

        def close(self) -> None:
            self.closed = True

    class DummyExternalBackend:
        def __init__(self) -> None:
            self.rows: list[ForexRateRecord] = []
            self.lme_rows: list[tuple[str, list[LmeRateRecord]]] = []
            self.checkpoints: dict[str, date] = {}

        def ensure_schema(self) -> None:
            return None

        def insert_rates(self, rows: list[ForexRateRecord]) -> PersistenceResult:
            self.rows.extend(rows)
            return PersistenceResult(inserted=len(rows))

        def insert_lme_rates(self, metal: str, rows: list[LmeRateRecord]) -> PersistenceResult:
            self.lme_rows.append((metal, rows))
            return PersistenceResult(inserted=len(rows))

        def update_ingestion_checkpoint(self, source: str, rate_date: date) -> None:
            self.checkpoints[source] = rate_date

    monkeypatch.setattr("fx_bharat.SQLiteBackend", DummySQLiteBackend)

    external = FxBharat(db_config="mysql://user:pwd@localhost:3306/fx")
    external._backend_strategy = DummyExternalBackend()

    start = date(2024, 1, 1)
    end = date(2024, 1, 31)
    external.migrate(from_date=start, to_date=end, chunk_size=50)

    backend = external._backend_strategy
    assert isinstance(backend, DummyExternalBackend)
    sqlite_backend = created_sqlite_backends[0]
    assert isinstance(sqlite_backend, DummySQLiteBackend)
    assert (start, end, "RBI") in sqlite_backend.fetch_calls
    assert (start, end, "SBI") in sqlite_backend.fetch_calls
    assert ("COPPER", start, end) in sqlite_backend.lme_calls
    assert ("ALUMINUM", start, end) in sqlite_backend.lme_calls
    assert len(backend.rows) == 2
    assert backend.checkpoints["RBI"] == start
    assert backend.checkpoints["SBI"] == start


def test_select_snapshot_dates_supports_all_frequencies() -> None:
    dates = [
        date(2024, 1, 1),
        date(2024, 1, 5),
        date(2024, 1, 12),
        date(2024, 2, 1),
        date(2025, 1, 1),
    ]

    weekly = FxBharat._select_snapshot_dates(dates, "weekly")
    monthly = FxBharat._select_snapshot_dates(dates, "monthly")
    yearly = FxBharat._select_snapshot_dates(dates, "yearly")

    assert len(weekly) == 4
    assert monthly == [date(2024, 1, 12), date(2024, 2, 1), date(2025, 1, 1)]
    assert yearly == [date(2024, 2, 1), date(2025, 1, 1)]



"""Data models shared across ingestion modules."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import date
from typing import Literal


@dataclass(slots=True)
class ForexRateRecord:
    """Representation of a single forex rate row extracted from RBI data."""

    rate_date: date
    currency: str
    rate: float
    source: str = "RBI"
    tt_buy: float | None = None
    tt_sell: float | None = None
    bill_buy: float | None = None
    bill_sell: float | None = None
    travel_card_buy: float | None = None
    travel_card_sell: float | None = None
    cn_buy: float | None = None
    cn_sell: float | None = None


@dataclass(slots=True)
class LmeRateRecord:
    """Representation of a single LME cash seller row."""

    rate_date: date
    price: float | None
    price_3_month: float | None
    stock: int | None
    metal: Literal["ALUMINUM", "COPPER"] | str = "COPPER"


__all__ = ["ForexRateRecord", "LmeRateRecord"]



from __future__ import annotations

from datetime import date

import pytest

from fx_bharat.db.sqlite_manager import SQLiteManager, _normalise_lme_metal
from fx_bharat.ingestion.models import ForexRateRecord, LmeRateRecord


def test_sqlite_manager_lme_update_and_filter(tmp_path) -> None:
    manager = SQLiteManager(tmp_path / "lme.db")
    try:
        first = LmeRateRecord(
            rate_date=date(2024, 2, 1),
            price=8500.0,
            price_3_month=8450.0,
            stock=100,
            metal="COPPER",
        )
        second = LmeRateRecord(
            rate_date=date(2024, 2, 1),
            price=8600.0,
            price_3_month=8550.0,
            stock=200,
            metal="COPPER",
        )
        manager.insert_lme_rates("COPPER", [first])
        result = manager.insert_lme_rates("COPPER", [second])
        assert result.updated == 1

        fetched = manager.fetch_lme_range("COPPER", start=date(2024, 2, 1), end=date(2024, 2, 1))
        assert fetched[0].stock == 200
    finally:
        manager.close()


def test_sqlite_manager_latest_rate_date(tmp_path) -> None:
    manager = SQLiteManager(tmp_path / "latest.db")
    try:
        assert manager.latest_rate_date("RBI") is None
    finally:
        manager.close()


def test_sqlite_manager_fetches_sbi_rows(tmp_path) -> None:
    manager = SQLiteManager(tmp_path / "sbi.db")
    try:
        manager.insert_rates(
            [
                ForexRateRecord(
                    rate_date=date(2024, 1, 2),
                    currency="USD",
                    rate=82.5,
                    source="SBI",
                    tt_buy=1.0,
                    tt_sell=2.0,
                    bill_buy=3.0,
                    bill_sell=4.0,
                    travel_card_buy=5.0,
                    travel_card_sell=6.0,
                    cn_buy=7.0,
                    cn_sell=8.0,
                )
            ]
        )
        rows = manager.fetch_range(source="SBI")
        assert rows[0].tt_sell == 2.0
    finally:
        manager.close()


def test_sqlite_manager_invalid_lme_metal(tmp_path) -> None:
    manager = SQLiteManager(tmp_path / "invalid.db")
    try:
        with pytest.raises(ValueError):
            manager.insert_lme_rates("GOLD", [])
    finally:
        manager.close()


def test_normalise_lme_metal_falls_back() -> None:
    assert _normalise_lme_metal("gold") == "GOLD"



from __future__ import annotations

from datetime import date
from pathlib import Path

from fx_bharat.seeds import populate_rbi_forex
from fx_bharat.seeds.populate_rbi_forex import PersistenceResult, RBINoReferenceRateError


class _DummyManager:
    def __init__(self) -> None:
        self.insert_calls: list = []

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False

    def ingestion_checkpoint(self, _source: str):
        return None

    def latest_rate_date(self, _source: str):
        return None

    def insert_rates(self, rows):
        self.insert_calls.append(rows)
        return PersistenceResult(inserted=len(rows), updated=0)

    def update_ingestion_checkpoint(
        self, _source: str, _day: date
    ) -> None:  # pragma: no cover - unused
        return None


class _DummyCSVParser:
    def parse(self, _path: Path):  # pragma: no cover - should never be called when RBI data missing
        raise AssertionError("parse should not be called")


class _DummyConverter:
    def to_csv(
        self, *_args, **_kwargs
    ):  # pragma: no cover - should never be called when RBI data missing
        raise AssertionError("to_csv should not be called")


class _DummyClient:
    download_dir = Path("/tmp")

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc, tb):
        return False

    def fetch_excel(self, *_args, **_kwargs):
        raise RBINoReferenceRateError("not ready")


def test_seed_rbi_forex_stops_on_missing_reference_rate(monkeypatch):
    monkeypatch.setattr(
        populate_rbi_forex, "SQLiteManager", lambda *_args, **_kwargs: _DummyManager()
    )
    monkeypatch.setattr(populate_rbi_forex, "RBISeleniumClient", lambda **_kwargs: _DummyClient())
    monkeypatch.setattr(populate_rbi_forex, "RBICSVParser", lambda: _DummyCSVParser())
    monkeypatch.setattr(populate_rbi_forex, "RBIWorkbookConverter", lambda: _DummyConverter())

    result = populate_rbi_forex.seed_rbi_forex("2025-11-20", "2025-11-21")

    assert result == PersistenceResult()



from __future__ import annotations

from datetime import date

import pytest

from fx_bharat.utils.date_range import (
    DateRange,
    _end_of_month,
    month_ranges,
    parse_date,
    split_ranges,
)


def test_parse_date_accepts_date_instance() -> None:
    today = date(2024, 1, 1)
    assert parse_date(today) == today


def test_month_ranges_rejects_invalid_order() -> None:
    with pytest.raises(ValueError):
        list(month_ranges("2024-02-01", "2024-01-01"))


def test_split_ranges_rejects_non_positive_window() -> None:
    with pytest.raises(ValueError):
        list(split_ranges("2024-01-01", "2024-01-10", 0))


def test_date_range_as_tuple() -> None:
    range_item = DateRange(start=date(2024, 1, 1), end=date(2024, 1, 31))
    assert range_item.as_tuple() == (date(2024, 1, 1), date(2024, 1, 31))


def test_end_of_month_december() -> None:
    assert _end_of_month(date(2024, 12, 15)) == date(2024, 12, 31)



from __future__ import annotations

from datetime import date
from pathlib import Path

from fx_bharat.db.sqlite_manager import SQLiteManager
from fx_bharat.ingestion.sbi_pdf import SBIPDFParser
from fx_bharat.seeds.populate_sbi_forex import seed_sbi_historical


def test_sbi_parser_extracts_rates(tmp_path: Path) -> None:
    pdf_path = tmp_path / "2024-01-02.pdf"
    pdf_path.write_text(
        """
        Forex Card Rates
        Date: 02/01/2024
        USD 83.11 84.11 83.01 84.21 82.91 84.31 82.71 84.51
        EURO 90.22 91.22 90.12 91.32 90.02 91.42 89.92 91.52
        STERLING 101.33 102.33 101.23 102.43 101.13 102.53 101.03 102.63
        """
    )
    parser = SBIPDFParser()

    parsed = parser.parse(pdf_path)

    assert parsed.rate_date == date(2024, 1, 2)
    rates = {row.currency: row.tt_buy for row in parsed.rates}
    assert rates == {"USD": 83.11, "EUR": 90.22, "GBP": 101.33}
    usd = next(row for row in parsed.rates if row.currency == "USD")
    assert (
        usd.tt_buy,
        usd.tt_sell,
        usd.bill_buy,
        usd.bill_sell,
        usd.travel_card_buy,
        usd.travel_card_sell,
        usd.cn_buy,
        usd.cn_sell,
    ) == (83.11, 84.11, 83.01, 84.21, 82.91, 84.31, 82.71, 84.51)
    assert {row.source for row in parsed.rates} == {"SBI"}


def test_seed_sbi_forex_populates_sqlite(tmp_path: Path) -> None:
    resource_dir = tmp_path / "resources"
    resource_dir.mkdir()
    pdf_path = resource_dir / "2024-01-05.pdf"
    pdf_path.write_text(
        """
        05/01/2024
        USD 83.5 84.5 83.4 84.6 83.3 84.7 83.2 84.8
        AUD 55.0 56.0 54.9 56.1 54.8 56.2 54.7 56.3
        """
    )
    db_path = tmp_path / "forex.db"

    result = seed_sbi_historical(
        db_path=db_path,
        resource_dir=resource_dir,
        start=date(2024, 1, 1),
        end=date(2024, 1, 31),
        download=False,
    )

    assert result.total == 2

    with SQLiteManager(db_path) as manager:
        rows = manager.fetch_range(source="SBI")
    assert {
        (
            row.rate_date,
            row.currency,
            row.tt_buy,
            row.tt_sell,
            row.bill_buy,
            row.bill_sell,
            row.travel_card_buy,
            row.travel_card_sell,
            row.cn_buy,
            row.cn_sell,
        )
        for row in rows
    } == {
        (
            date(2024, 1, 5),
            "USD",
            83.5,
            84.5,
            83.4,
            84.6,
            83.3,
            84.7,
            83.2,
            84.8,
        ),
        (
            date(2024, 1, 5),
            "AUD",
            55.0,
            56.0,
            54.9,
            56.1,
            54.8,
            56.2,
            54.7,
            56.3,
        ),
    }


def test_parser_reads_decimal_spacing_and_cn_from_pdf() -> None:
    parser = SBIPDFParser()
    pdf_path = Path("resources/2025/1/2025-01-01.pdf")

    parsed = parser.parse(pdf_path)

    usd_rows = [row for row in parsed.rates if row.currency == "USD"]
    assert len(usd_rows) == 1
    usd = usd_rows[0]
    assert parsed.rate_date == date(2025, 1, 1)
    assert usd.tt_buy == 85.22
    assert usd.tt_sell == 86.07
    assert usd.bill_buy == 85.15
    assert usd.bill_sell == 86.24
    assert usd.travel_card_buy == 85.15
    assert usd.travel_card_sell == 86.24
    assert usd.cn_buy == 84.15
    assert usd.cn_sell == 86.55


def test_sbi_parser_infers_date_from_filename_when_missing_date(tmp_path: Path) -> None:
    pdf_path = tmp_path / "2024-12-31.pdf"
    pdf_path.write_text("USD 80 81 82 83 84 85 86 87")

    parsed = SBIPDFParser().parse(pdf_path)

    assert parsed.rate_date == date(2024, 12, 31)
    usd = parsed.rates[0]
    assert usd.currency == "USD"
    assert (
        usd.tt_buy,
        usd.tt_sell,
        usd.bill_buy,
        usd.bill_sell,
        usd.travel_card_buy,
        usd.travel_card_sell,
        usd.cn_buy,
        usd.cn_sell,
    ) == (80, 81, 82, 83, 84, 85, 86, 87)


def test_sbi_parser_deduplicates_currency_rows(tmp_path: Path) -> None:
    pdf_path = tmp_path / "2024-02-01.pdf"
    pdf_path.write_text(
        """
        Date: 01/02/2024
        USD 83.11 84.11 83.01 84.21 82.91 84.31 82.71 84.51
        USD 90.00 90.00 90.00 90.00 90.00 90.00 90.00 90.00
        """
    )

    parsed = SBIPDFParser().parse(pdf_path)

    usd_rows = [row for row in parsed.rates if row.currency == "USD"]
    assert len(usd_rows) == 1
    usd = usd_rows[0]
    assert usd.rate_date == date(2024, 2, 1)
    assert usd.tt_buy == 83.11


def test_sbi_parser_falls_back_to_filename_when_header_date_invalid(tmp_path: Path) -> None:
    pdf_path = tmp_path / "2020-10-12.pdf"
    pdf_path.write_text(
        """
        Date: 20/20/2020
        USD 80 81 82 83 84 85 86 87
        """
    )

    parsed = SBIPDFParser().parse(pdf_path)

    assert parsed.rate_date == date(2020, 10, 12)
    usd = parsed.rates[0]
    assert usd.currency == "USD"
    assert usd.tt_buy == 80



from __future__ import annotations

import pytest

import fx_bharat.seeds as seeds


def test_seeds_module_lazy_getattr_exports() -> None:
    assert callable(getattr(seeds, "seed_rbi_forex"))
    assert callable(getattr(seeds, "seed_lme_prices"))
    assert callable(getattr(seeds, "seed_sbi_today"))


def test_seeds_module_invalid_attribute() -> None:
    with pytest.raises(AttributeError):
        getattr(seeds, "unknown_seed")



"""SQLite backend strategy implementation."""

from __future__ import annotations

from datetime import date
from pathlib import Path
from typing import Sequence

from fx_bharat.db import DEFAULT_SQLITE_DB_PATH
from fx_bharat.db.base_backend import BackendStrategy
from fx_bharat.db.sqlite_manager import PersistenceResult, SQLiteManager
from fx_bharat.ingestion.models import ForexRateRecord, LmeRateRecord


class SQLiteBackend(BackendStrategy):
    """Backend strategy that stores rates in the bundled SQLite database."""

    def __init__(
        self,
        db_path: str | Path = DEFAULT_SQLITE_DB_PATH,
        *,
        manager: SQLiteManager | None = None,
    ) -> None:
        self.manager = manager or SQLiteManager(db_path)
        self.db_path = Path(self.manager.db_path)

    def ensure_schema(self) -> None:
        # ``SQLiteManager`` creates the schema in its constructor, so nothing else
        # is required here.
        return None

    def insert_rates(self, rows: Sequence[ForexRateRecord]) -> PersistenceResult:
        return self.manager.insert_rates(rows)

    def fetch_range(
        self,
        start: date | None = None,
        end: date | None = None,
        *,
        source: str | None = None,
    ) -> list[ForexRateRecord]:
        return self.manager.fetch_range(start, end, source=source)

    def insert_lme_rates(self, metal: str, rows: Sequence[LmeRateRecord]) -> PersistenceResult:
        return self.manager.insert_lme_rates(metal, rows)

    def fetch_lme_range(
        self, metal: str, start: date | None = None, end: date | None = None
    ) -> list[LmeRateRecord]:
        return self.manager.fetch_lme_range(metal, start, end)

    def update_ingestion_checkpoint(self, source: str, rate_date: date) -> None:
        self.manager.update_ingestion_checkpoint(source, rate_date)

    def close(self) -> None:  # pragma: no cover - trivial delegator
        self.manager.close()


__all__ = ["SQLiteBackend"]



from __future__ import annotations

from datetime import date
from pathlib import Path

from fx_bharat.ingestion.strategy import IngestionStrategy


class _DummyStrategy:
    def fetch(self, start_date: date, end_date: date, *, destination: Path | None = None) -> Path:
        target = destination or Path("default.dat")
        target.write_text(f"{start_date.isoformat()}-{end_date.isoformat()}")
        return target


def test_ingestion_strategy_contract(tmp_path: Path) -> None:
    strategy: IngestionStrategy = _DummyStrategy()
    dest = tmp_path / "artifact.dat"

    result = strategy.fetch(date(2024, 1, 1), date(2024, 1, 2), destination=dest)

    assert result == dest
    assert result.read_text() == "2024-01-01-2024-01-02"
